import codecs
import itertools
import logging
import os
import re
import shlex
import signal
import six
import traceback

from collections import defaultdict
from glob import glob
from subprocess import call

from insights.core import blacklist, dr, filters
from insights.core.context import ExecutionContext, FSRoots, HostContext
from insights.core.exceptions import (
    BlacklistedSpec,
    ContentException,
    NoFilterException,
    SkipComponent,
)
from insights.core.plugins import component, datasource, is_datasource
from insights.core.serde import deserializer, serializer
from insights.util import fs, streams, which
from insights.util import deprecated
from insights.util.mangle import mangle_command

log = logging.getLogger(__name__)


SAFE_ENV = {
    "PATH": os.path.pathsep.join(
        [
            "/bin",
            "/usr/bin",
            "/sbin",
            "/usr/sbin",
            "/usr/share/Modules/bin",
        ]
    ),
    "LC_ALL": "C",
}
"""
A minimal set of environment variables for use in subprocess calls
"""
if "LANG" in os.environ:
    SAFE_ENV["LANG"] = os.environ["LANG"]

safe_open, encoding = (open, "utf-8") if six.PY3 else (codecs.open, None)


class ContentProvider(object):
    def __init__(self):
        self.cmd = None
        self.args = None
        self.rc = None
        self.root = None
        self.relative_path = None
        self.save_as = None
        self.loaded = False
        self._content = None
        self._exception = None

    def load(self):
        raise NotImplementedError()

    def stream(self):
        """
        Returns a generator of lines instead of a list of lines.
        """
        st = self._stream()
        for l in next(st):
            yield l.rstrip("\n")

    def _stream(self):
        raise NotImplementedError()

    def _clean_content(self):
        """
        Clean (Obfuscate and Redact) the Spec Content ONLY when doing
        collection.
        """
        content = self.content  # load first for debugging info order
        if isinstance(self.ctx, HostContext) and self.ds and self.cleaner:
            cleans = []
            # Redacting?
            no_red = getattr(self.ds, 'no_redact', False)
            cleans.append("Redact") if not no_red else None
            # Obfuscating?
            no_obf = getattr(self.ds, 'no_obfuscate', [])
            cleans.append("Obfuscate") if len(no_obf) < 2 else None
            # Cleaning - Entry
            if cleans:
                log.debug("Cleaning (%s) %s", "/".join(cleans), "/" + self.relative_path)
                content = self.cleaner.clean_content(
                    content,
                    obf_funcs=self.cleaner.get_obfuscate_functions(self.relative_path, no_obf),
                    no_redact=no_red,
                )
                if len(content) == 0:
                    log.debug("Skipping %s due to empty after cleaning", self.path)
                    raise ContentException("Empty after cleaning: %s" % self.path)
            else:
                log.debug("Skipping cleaning %s", "/" + self.relative_path)
        return content

    @property
    def path(self):
        return os.path.join(self.root, self.relative_path)

    @property
    def content(self):
        if self._exception:
            raise self._exception

        if self._content is None and not self.loaded:
            try:
                self._content = self.load()
            except Exception as ex:
                self._exception = ex
                raise

        if len(self._content) == 0:
            log.debug("File is empty (after filtering): %s", self.path)
            if isinstance(self.ctx, HostContext):
                # Do not collect empty spec
                raise ContentException("Empty (after filtering): %s" % self.path)

        return self._content

    def write(self, dst):
        fs.ensure_path(os.path.dirname(dst))
        # Clean Spec Content when writing it down to disk before uploading
        content = "\n".join(self._clean_content())
        content = content.encode("utf-8") if six.PY3 else content
        with open(dst, "wb") as f:
            f.write(content)

        self.loaded = False

    def __repr__(self):
        msg = "<%s(path=%r, cmd=%r)>"
        return msg % (self.__class__.__name__, self.path or "", self.cmd or "")

    def __unicode__(self):
        return self.__repr__()

    def __str__(self):
        return self.__unicode__()


class DatasourceProvider(ContentProvider):
    def __init__(
        self,
        content,
        relative_path,
        root='/',
        save_as=None,
        ds=None,
        ctx=None,
        cleaner=None,
        no_obfuscate=None,
        no_redact=False,
    ):
        super(DatasourceProvider, self).__init__()
        self.relative_path = relative_path.lstrip("/")
        self.save_as = save_as
        self._content = content if isinstance(content, list) else content.splitlines()
        self.root = root
        self.ds = ds or self
        self.ctx = ctx
        self.cleaner = cleaner
        self.no_obfuscate = no_obfuscate or []
        self.no_redact = no_redact

    def _stream(self):
        """
        Returns a generator of lines instead of a list of lines.
        """
        yield self._content

    def load(self):
        self.loaded = True
        return self.content


class FileProvider(ContentProvider):
    def __init__(self, relative_path, root="/", save_as=None, ds=None, ctx=None, cleaner=None):
        super(FileProvider, self).__init__()
        self.ds = ds
        self.ctx = ctx
        self.root = root
        self.cleaner = cleaner
        self.relative_path = relative_path.lstrip("/")
        self.save_as = save_as
        self.file_name = os.path.basename(self.path)

        self.validate()

    def validate(self):
        # 1. No Such File
        if not os.path.exists(self.path):
            raise ContentException("%s does not exist." % self.path)
        # 2. Check only when collecting
        if isinstance(self.ctx, HostContext):
            # 2.1 No Filters for 'filterable=True' Specs
            if (
                self.ds
                and filters.ENABLED
                and any(s.filterable for s in dr.get_registry_points(self.ds))
                and not filters.get_filters(self.ds)
            ):
                raise NoFilterException("Skipping %s due to no filters." % dr.get_name(self.ds))
            # 2.2 Customer Prohibits Collection
            if not blacklist.allow_file("/" + self.relative_path):
                log.warning("WARNING: Skipping file %s", "/" + self.relative_path)
                raise BlacklistedSpec()

        resolved = os.path.realpath(self.path)
        if not resolved.startswith(os.path.realpath(self.root)):
            msg = "Relative path points outside the root: %s -> %s."
            raise Exception(msg % (self.path, resolved))

        if not os.access(self.path, os.R_OK):
            raise ContentException("Cannot access %s" % self.path)

    def __repr__(self):
        return '%s("%r")' % (self.__class__.__name__, self.path)


class MetadataProvider(FileProvider):
    """
    .. warning::
        This Class is deprecated and will be removed from 3.5.0.
        Please collect built-in file by using datasource spec directly, see
        :mod:`insights.specs.datasources.client_metadata`.

    Class used for insights-core built-in files.  These files should not
    be filtered, redacted or blocked.
    """

    def __init__(self, relative_path, root="/", save_as=None, ds=None, ctx=None, cleaner=None):
        deprecated(
            MetadataProvider,
            "Please collect the built-in file via datasource spec instead.",
            "3.5.0",
        )
        super(MetadataProvider, self).__init__(relative_path, root, save_as, ds, ctx, cleaner)

    def _stream(self):
        """
        Returns a generator of lines instead of a list of lines.
        """
        if self._exception:
            raise self._exception
        try:
            with safe_open(self.path, "r", encoding=encoding, errors="surrogateescape") as f:
                yield f
        except StopIteration:
            raise
        except Exception as ex:
            self._exception = ex
            raise ContentException(str(ex))

    def validate(self):
        # Validate built-in metedata files only when insights-run
        if not isinstance(self.ctx, HostContext):
            super(MetadataProvider, self).validate()
        # But DO NOT validate them when core-collecting

    def load(self):
        self.loaded = True
        with safe_open(self.path, "r", encoding=encoding, errors="surrogateescape") as f:
            return [l.rstrip("\n") for l in f]


class RawFileProvider(FileProvider):
    """
    Class used in datasources that returns the contents of a file a single
    string. The file is not filtered/obfuscated/redacted.
    """

    def load(self):
        self.loaded = True
        with open(self.path, 'rb') as f:
            return f.read()

    def write(self, dst):
        fs.ensure_path(os.path.dirname(dst))
        call([which("cp", env=SAFE_ENV), self.path, dst], env=SAFE_ENV)


class TextFileProvider(FileProvider):
    """
    Class used in datasources that returns the contents of a file a list of
    lines. Each line is filtered if filters are defined for the datasource.
    """

    def create_args(self):
        args = []
        _filters = "\n".join(filters.get_filters(self.ds))
        if _filters:
            args.append(["grep", "-F", _filters, self.path])

        return args

    def load(self):
        self.loaded = True
        args = self.create_args()
        if args:
            rc, out = self.ctx.shell_out(args, keep_rc=True, env=SAFE_ENV)
            self.rc = rc
            return out

        with safe_open(self.path, "r", encoding=encoding, errors="surrogateescape") as f:
            return [l.rstrip("\n") for l in f]

    def _stream(self):
        """
        Returns a generator of lines instead of a list of lines.
        """
        if self._exception:
            raise self._exception
        try:
            if self._content:
                yield self._content
            else:
                args = self.create_args()
                if args:
                    with streams.connect(*args, env=SAFE_ENV) as s:
                        yield s
                else:
                    with safe_open(
                        self.path, "r", encoding=encoding, errors="surrogateescape"
                    ) as f:
                        yield f
        except StopIteration:
            raise
        except Exception as ex:
            self._exception = ex
            raise ContentException(str(ex))


class SerializedOutputProvider(TextFileProvider):
    pass


class SerializedRawOutputProvider(RawFileProvider):
    pass


class CommandOutputProvider(ContentProvider):
    """
    Class used in datasources to return output from commands.
    """

    def __init__(
        self,
        cmd,
        ctx,
        root="insights_commands",
        save_as=None,
        args=None,
        split=True,
        keep_rc=False,
        ds=None,
        timeout=None,
        inherit_env=None,
        override_env=None,
        signum=None,
        cleaner=None,
    ):
        super(CommandOutputProvider, self).__init__()
        self.cmd = cmd if six.PY3 else str(cmd)
        self.root = root
        self.save_as = save_as
        self.ctx = ctx
        self.args = args  # already interpolated into cmd - stored here for context.
        self.split = split
        self.keep_rc = keep_rc
        self.ds = ds
        self.timeout = timeout
        self.inherit_env = inherit_env if inherit_env is not None else []
        self.override_env = override_env if override_env is not None else dict()
        self.signum = signum or signal.SIGKILL
        self.rc = None
        self.cleaner = cleaner

        self._misc_settings()
        self._content = None
        self._env = self.create_env()

        self.validate()

    def _misc_settings(self):
        """Re-implement it according to the actual case of the sub-class"""
        self.relative_path = mangle_command(self.cmd)

    def validate(self):
        # 1. No Such Command
        cmd = shlex.split(self.cmd)[0]
        if not which(cmd, env=self._env):
            raise ContentException("Command not found: %s" % cmd)
        # 2. Check only when collecting
        if isinstance(self.ctx, HostContext):
            # 2.1 No Filters for 'filterable=True' Specs
            if (
                self.ds
                and filters.ENABLED
                and any(s.filterable for s in dr.get_registry_points(self.ds))
                and not filters.get_filters(self.ds)
            ):
                raise NoFilterException("Skipping %s due to no filters." % dr.get_name(self.ds))
            # 2.2 Customer Prohibits Collection
            if not blacklist.allow_command(self.cmd):
                log.warning("WARNING: Skipping command %s", self.cmd)
                raise BlacklistedSpec()

    def create_args(self):
        command = [shlex.split(self.cmd)]

        if self.split:
            _filters = "\n".join(filters.get_filters(self.ds))
            if _filters:
                command.append(["grep", "-F", _filters])

        return command

    def create_env(self):
        env = dict(SAFE_ENV)

        for e in self.inherit_env:
            if e in os.environ:
                env[e] = os.environ[e]

        for k, v in self.override_env.items():
            env[k] = v

        return env

    def load(self):
        command = self.create_args()

        raw = self.ctx.shell_out(
            command,
            split=self.split,
            keep_rc=self.keep_rc,
            timeout=self.timeout,
            env=self._env,
            signum=self.signum,
        )
        if self.keep_rc:
            self.rc, output = raw
        else:
            output = raw
        return output

    def _stream(self):
        """
        Returns a generator of lines instead of a list of lines.
        """
        if self._exception:
            raise self._exception
        try:
            if self._content:
                yield self._content
            else:
                command = self.create_args()
                with self.ctx.connect(*command, env=self._env, timeout=self.timeout) as s:
                    yield s
        except StopIteration:
            raise
        except Exception as ex:
            self._exception = ex
            raise ContentException(str(ex))

    def __repr__(self):
        return 'CommandOutputProvider("%r")' % self.cmd


class ContainerProvider(CommandOutputProvider):
    def __init__(
        self,
        cmd_path,
        ctx,
        image=None,
        args=None,
        split=True,
        keep_rc=False,
        ds=None,
        timeout=None,
        inherit_env=None,
        override_env=None,
        signum=None,
        cleaner=None,
    ):
        # cmd  = "<podman|docker> exec container_id command"
        # path = "<podman|docker> exec container_id cat path"
        self.image = image
        super(ContainerProvider, self).__init__(
            cmd_path,
            ctx,
            "insights_containers",
            None,
            args,
            split,
            keep_rc,
            ds,
            timeout,
            inherit_env,
            override_env,
            signum,
            cleaner,
        )


class ContainerFileProvider(ContainerProvider):
    def _misc_settings(self):
        engine, _, container_id, _, path = self.cmd.split(None, 4)
        self.engine = os.path.basename(engine)
        self.container_id = container_id
        self.relative_path = os.path.join(container_id, path.lstrip('/'))

    def __repr__(self):
        return 'ContainerFileProvider("%r")' % self.cmd


class ContainerCommandProvider(ContainerProvider):
    def _misc_settings(self):
        engine, _, container_id, cmd = self.cmd.split(None, 3)
        self.engine = os.path.basename(engine)
        self.container_id = container_id
        self.relative_path = os.path.join(container_id, "insights_commands", mangle_command(cmd))

    def __repr__(self):
        return 'ContainerCommandProvider("%r")' % self.cmd


class RegistryPoint(object):
    # Marker class for declaring that an element of a `SpecSet` subclass
    # is a registry point against which further subclasses can register
    # datasource implementations by simply declaring them with the same name.
    #
    # intentionally not a docstring so this doesn't show up in pydoc.
    def __init__(
        self,
        metadata=None,
        multi_output=False,
        raw=False,
        filterable=False,
        no_obfuscate=None,
        no_redact=False,
        prio=0,
    ):
        self.metadata = metadata
        self.multi_output = multi_output
        self.no_obfuscate = [] if no_obfuscate is None else no_obfuscate
        self.no_redact = no_redact
        self.prio = prio
        self.raw = raw
        self.filterable = filterable
        self.__name__ = self.__class__.__name__
        datasource(
            [],
            metadata=metadata,
            multi_output=multi_output,
            raw=raw,
            filterable=filterable,
            no_obfuscate=self.no_obfuscate,
            no_redact=no_redact,
            prio=prio,
        )(self)

    def __call__(self, broker):
        for c in reversed(dr.get_delegate(self).deps):
            if c in broker:
                return broker[c]
        raise SkipComponent()

    def __repr__(self):
        return dr.get_name(self)


class SpecDescriptor(object):
    # Descriptor Protocol handler that returns the literal function from a
    # class during dot (.) access.
    #
    # intentionally not a docstring so this doesn't show up in pydoc.
    def __init__(self, func):
        self.func = func

    def __get__(self, obj, obj_type):
        return self.func

    def __set__(self, obj, val):
        raise AttributeError()


def _get_ctx_dependencies(component):
    ctxs = set()
    for c in dr.walk_tree(component):
        try:
            if issubclass(c, ExecutionContext):
                ctxs.add(c)
        except Exception:
            pass
    return ctxs


def _register_context_handler(parents, component):
    name = component.__name__
    parents = list(itertools.takewhile(lambda x: name in x.registry, parents))
    if not parents:
        return

    # If the new component handles a context, we need to tell the
    # previously registered components that would have handled it to ignore it.

    # The components that handle a context are registered on the highest class
    # of the MRO list. This is so overrides work correctly even if a
    # component isn't a direct sibling of the component it's overriding.

    # instead of trying to unhook all of the dependencies, we just tell the
    # previous handler of a context to ignore it.
    ctx_handlers = parents[-1].context_handlers
    for c in _get_ctx_dependencies(component):
        for old in ctx_handlers[name][c]:
            dr.add_ignore(old, c)
        ctx_handlers[name][c].append(component)


def _resolve_registry_points(cls, base, dct):
    module = cls.__module__
    parents = [x for x in cls.__mro__ if x not in (cls, SpecSet, object)]

    for k, v in dct.items():
        if isinstance(v, RegistryPoint):
            # add v under its name to this class's registry.
            v.__name__ = k
            cls.registry[k] = v

        if is_datasource(v):
            v.__qualname__ = ".".join([cls.__name__, k])
            v.__name__ = k
            v.__module__ = module
            setattr(cls, k, SpecDescriptor(v))
            if k in base.registry:
                # if the datasource has the same name as a RegistryPoint in the
                # base class, the datasource to the RegistryPoint.
                point = base.registry[k]

                # TODO: log when RegistryPoint and implementation properties
                # TODO: aren't the same.
                delegate = dr.get_delegate(v)
                v.filterable = delegate.filterable = point.filterable
                v.raw = delegate.raw = point.raw
                v.multi_output = delegate.multi_output = point.multi_output
                v.no_obfuscate = delegate.no_obfuscate = point.no_obfuscate
                v.no_redact = delegate.no_redact = point.no_redact
                v.prio = delegate.prio = point.prio

                # the RegistryPoint gets the implementation datasource as a
                # dependency
                dr.add_dependency(point, v)

                # Datasources override previously defined datasources of the
                # same name for contexts they all depend on. Here we tell
                # datasources of the same name not to execute under contexts
                # the new datasource will handle.
                _register_context_handler(parents, v)


class SpecSetMeta(type):
    """
    The metaclass that converts RegistryPoint markers to registry point
    datasources and hooks implementations for them into the registry.
    """

    def __new__(cls, name, bases, dct):
        dct["context_handlers"] = defaultdict(lambda: defaultdict(list))
        dct["registry"] = {}
        return super(SpecSetMeta, cls).__new__(cls, name, bases, dct)

    def __init__(cls, name, bases, dct):
        if name == "SpecSet":
            return
        if len(bases) > 1:
            raise Exception("SpecSet subclasses must inherit from only one class.")
        _resolve_registry_points(cls, bases[0], dct)


class SpecSet(six.with_metaclass(SpecSetMeta)):
    """
    The base class for all spec declarations. Extend this class and define your
    datasources directly or with a `SpecFactory`.
    """

    pass


def _get_context(context, broker):
    if isinstance(context, list):
        return dr.first_of(context, broker)
    return broker.get(context)


class simple_file(object):
    """
    Creates a datasource that reads the file at path when evaluated.

    Args:
        path (str): path to the file to collect.
        save_as (str or None): path to save the collected file as.
            - It should be a relative path and any starting '/' will be removed.
            - If it's a path which ending with '/', the collected file will be
            stored to the "save_as" directory,
            - If it's a path which not ending with '/', the collected file will
            be renamed to the file with "save_as" as the full path.
        context (ExecutionContext): the context under which the datasource
            should run.
        kind (FileProvider): One of TextFileProvider or RawFileProvider.

    Returns:
        function: A datasource that reads all files matching the glob patterns.
    """

    def __init__(
        self, path, save_as=None, context=None, deps=None, kind=TextFileProvider, **kwargs
    ):
        deps = deps if deps is not None else []
        self.path = path
        self.save_as = save_as.lstrip("/") if save_as else None
        self.context = context or FSRoots
        self.kind = kind
        self.raw = kind is RawFileProvider
        self.__name__ = self.__class__.__name__
        datasource(self.context, *deps, raw=self.raw, **kwargs)(self)

    def __call__(self, broker):
        ctx = _get_context(self.context, broker)
        cleaner = broker.get('cleaner')
        return self.kind(
            ctx.locate_path(self.path),
            root=ctx.root,
            save_as=self.save_as,
            ds=self,
            ctx=ctx,
            cleaner=cleaner,
        )


class glob_file(object):
    """
    Creates a datasource that reads all files matching the glob pattern(s).

    Args:
        patterns (str or [str]): glob pattern(s) of paths to collect.
        save_as (str or None): directory path to save the collected files as.
            - It should be a relative path and any starting '/' will be removed
            and an ending '/' will be added.
        ignore (regex): a regular expression that is used to filter the paths
            matched by pattern(s).
        context (ExecutionContext): the context under which the datasource
            should run.
        kind (FileProvider): One of TextFileProvider or RawFileProvider.
        max_files (int): Maximum number of glob files to process.

    Returns:
        function: A datasource that reads all files matching the glob patterns.
    """

    def __init__(
        self,
        patterns,
        save_as=None,
        ignore=None,
        context=None,
        deps=None,
        kind=TextFileProvider,
        max_files=1000,
        **kwargs
    ):
        deps = deps if deps is not None else []
        if not isinstance(patterns, (list, set)):
            patterns = [patterns]
        self.patterns = patterns
        self.ignore = ignore
        self.save_as = os.path.join(save_as.lstrip("/"), '') if save_as else None
        self.ignore_func = re.compile(ignore).search if ignore else lambda x: False
        self.context = context or FSRoots
        self.kind = kind
        self.raw = kind is RawFileProvider
        self.max_files = max_files
        self.__name__ = self.__class__.__name__
        datasource(self.context, *deps, multi_output=True, raw=self.raw, **kwargs)(self)

    def __call__(self, broker):
        cleaner = broker.get('cleaner')
        ctx = _get_context(self.context, broker)
        root = ctx.root
        results = []
        for pattern in self.patterns:
            pattern = ctx.locate_path(pattern)
            for path in sorted(glob(os.path.join(root, pattern.lstrip('/')))):
                if self.ignore_func(path) or os.path.isdir(path):
                    continue
                try:
                    results.append(
                        self.kind(
                            path[len(root) :],
                            root=root,
                            save_as=self.save_as,
                            ds=self,
                            ctx=ctx,
                            cleaner=cleaner,
                        )
                    )
                except NoFilterException as nfe:
                    raise nfe
                except Exception:
                    log.debug(traceback.format_exc())
        if results:
            if len(results) > self.max_files:
                raise ContentException(
                    "Number of files returned [{0}] is over the {1} file limit, please refine "
                    "the specs file pattern to narrow down results".format(
                        len(results), self.max_files
                    )
                )
            return results
        raise ContentException("[%s] didn't match." % ', '.join(self.patterns))


class head(object):
    """
    Return the first element of any datasource that produces a list.
    """

    def __init__(self, dep, **kwargs):
        self.dep = dep
        self.__name__ = self.__class__.__name__
        datasource(dep, **kwargs)(self)

    def __call__(self, lst):
        c = lst[self.dep]
        if lst:
            return c[0]
        raise SkipComponent()


class first_file(object):
    """
    Creates a datasource that returns the first existing and readable file in
    files.

    Args:
        paths (str): list of paths to find and collect.
        save_as (str or None): path to save the collected file as.
            - It should be a relative path and any starting '/' will be removed.
            - If it's a path which ending with '/', the collected file will be
            stored to the "save_as" directory,
            - If it's a path which not ending with '/', the collected file will
            be renamed to the file with "save_as" as the full path.
        context (ExecutionContext): the context under which the datasource
            should run.
        kind (FileProvider): One of TextFileProvider or RawFileProvider.

    Returns:
        function: A datasource that returns the first file in files that exists
            and is readable
    """

    def __init__(
        self, paths, save_as=None, context=None, deps=None, kind=TextFileProvider, **kwargs
    ):
        deps = deps if deps is not None else []
        self.paths = paths
        self.save_as = save_as.lstrip("/") if save_as else None
        self.context = context or FSRoots
        self.kind = kind
        self.raw = kind is RawFileProvider
        self.__name__ = self.__class__.__name__
        datasource(self.context, *deps, raw=self.raw, **kwargs)(self)

    def __call__(self, broker):
        cleaner = broker.get('cleaner')
        ctx = _get_context(self.context, broker)
        root = ctx.root
        for p in self.paths:
            try:
                return self.kind(
                    ctx.locate_path(p),
                    root=root,
                    save_as=self.save_as,
                    ds=self,
                    ctx=ctx,
                    cleaner=cleaner,
                )
            except NoFilterException as nfe:
                raise nfe
            except Exception:
                pass
        raise ContentException("None of [%s] found." % ', '.join(self.paths))


class listdir(object):
    """
    Execute a simple directory listing of all the files and directories in
    path.

    Args:
        path (str): directory to list.
        context (ExecutionContext): the context under which the datasource
            should run.
        ignore (str): regular expression defining names to ignore.

    Returns:
        function: A datasource that returns a sorted list of file and directory
            names in the directory specified by path. The list will be empty when
            the directory is empty or all names get ignored.
    """

    def __init__(self, path, context=None, ignore=None, deps=None):
        deps = deps if deps is not None else []
        self.path = path
        self.context = context or FSRoots
        self.ignore = ignore
        self.ignore_func = re.compile(ignore).search if ignore else lambda x: False
        self.__name__ = self.__class__.__name__
        datasource(self.context, *deps)(self)

    def __call__(self, broker):
        ctx = _get_context(self.context, broker)
        p = os.path.join(ctx.root, self.path.lstrip('/'))
        p = ctx.locate_path(p)
        try:
            result = os.listdir(p)
        except OSError as e:
            raise ContentException(str(e))
        return sorted([r for r in result if not self.ignore_func(r)])


class listglob(listdir):
    """
    List paths matching a glob pattern.

    Args:
        pattern (str): glob pattern to list.
        context (ExecutionContext): the context under which the datasource
            should run.
        ignore (str): regular expression defining paths to ignore.

    Returns:
        function: A datasource that returns the list of paths that match
            the given glob pattern. The list will be empty when nothing matches.
    """

    def __call__(self, broker):
        ctx = _get_context(self.context, broker)
        p = os.path.join(ctx.root, self.path.lstrip('/'))
        p = ctx.locate_path(p)
        result = glob(p)
        # generator expression; we don't need the full list at this step
        result = (os.path.relpath(r, start=ctx.root) for r in result)
        result = sorted([r for r in result if not self.ignore_func(r)])
        return result


class simple_command(object):
    """
    Execute a simple command that has no dynamic arguments

    Args:
        cmd (str): the command(s) to execute. Breaking apart a command
            string that might contain multiple commands separated by a pipe,
            getting them ready for subproc operations.
            IE. A command with filters applied
        save_as (str or None): path to save the collected file as.
            - It should be a relative path in which any starting and ending
            '/' will be removed, the collected file will be renamed to
            `save_as` under the 'insights_commands' directory.
        context (ExecutionContext): the context under which the datasource
            should run.
        split (bool): whether the output of the command should be split into a
            list of lines
        keep_rc (bool): whether to return the error code returned by the
            process executing the command. If False, any return code other than
            zero with raise a CalledProcessError. If True, the return code and
            output are always returned.
        timeout (int): Number of seconds to wait for the command to complete.
            If the timeout is reached before the command returns, a
            CalledProcessError is raised. If None, timeout is infinite.
        inherit_env (list): The list of environment variables to inherit from the
            calling process when the command is invoked.
        override_env (dict): A dict of environment variables to override from the
            calling process when the command is invoked.

    Returns:
        function: A datasource that returns the output of a command that takes
            no arguments
    """

    def __init__(
        self,
        cmd,
        save_as=None,
        context=HostContext,
        deps=None,
        split=True,
        keep_rc=False,
        timeout=None,
        inherit_env=None,
        override_env=None,
        signum=None,
        **kwargs
    ):
        deps = deps if deps is not None else []
        self.cmd = cmd
        self.context = context
        self.save_as = save_as.strip("/") if save_as else None  # strip as a relative file path
        self.split = split
        self.raw = not split
        self.keep_rc = keep_rc
        self.timeout = timeout
        self.inherit_env = inherit_env if inherit_env is not None else []
        self.override_env = override_env if override_env is not None else dict()
        self.signum = signum
        self.__name__ = self.__class__.__name__
        datasource(self.context, *deps, raw=self.raw, **kwargs)(self)

    def __call__(self, broker):
        cleaner = broker.get('cleaner')
        ctx = broker[self.context]
        return CommandOutputProvider(
            self.cmd,
            ctx,
            save_as=self.save_as,
            split=self.split,
            keep_rc=self.keep_rc,
            ds=self,
            timeout=self.timeout,
            inherit_env=self.inherit_env,
            override_env=self.override_env,
            signum=self.signum,
            cleaner=cleaner,
        )


class command_with_args(object):
    """
    Execute a command that has dynamic arguments

    Args:
        cmd (str): the command to execute. Breaking apart a command
            string that might require arguments.
        provider (str or tuple): argument string or a tuple of argument strings.
        save_as (str or None): path to save the collected file as.
            - It should be a relative path in which any starting and ending
            '/' will be removed, the collected file will be renamed to
            `save_as` under the 'insights_commands' directory.
        context (ExecutionContext): the context under which the datasource
            should run.
        split (bool): whether the output of the command should be split into a
            list of lines
        keep_rc (bool): whether to return the error code returned by the
            process executing the command. If False, any return code other than
            zero with raise a CalledProcessError. If True, the return code and
            output are always returned.
        timeout (int): Number of seconds to wait for the command to complete.
            If the timeout is reached before the command returns, a
            CalledProcessError is raised. If None, timeout is infinite.
        inherit_env (list): The list of environment variables to inherit from the
            calling process when the command is invoked.
        override_env (dict): A dict of environment variables to override from the
            calling process when the command is invoked.

    Returns:
        function: A datasource that returns the output of a command that takes
            specified arguments passed by the provider.
    """

    def __init__(
        self,
        cmd,
        provider,
        save_as=None,
        context=HostContext,
        deps=None,
        split=True,
        keep_rc=False,
        timeout=None,
        inherit_env=None,
        override_env=None,
        signum=None,
        **kwargs
    ):
        deps = deps if deps is not None else []
        self.cmd = cmd if six.PY3 else str(cmd)
        self.provider = provider
        self.save_as = save_as.strip("/") if save_as else None  # strip as a relative file path
        self.context = context
        self.split = split
        self.raw = not split
        self.keep_rc = keep_rc
        self.timeout = timeout
        self.inherit_env = inherit_env if inherit_env is not None else []
        self.override_env = override_env if override_env is not None else dict()
        self.signum = signum
        self.__name__ = self.__class__.__name__
        datasource(self.provider, self.context, *deps, raw=self.raw, **kwargs)(self)

    def __call__(self, broker):
        cleaner = broker.get('cleaner')
        source = broker[self.provider]
        ctx = broker[self.context]
        if isinstance(source, ContentProvider):
            source = source.content
        if not isinstance(source, (six.text_type, str, tuple)):
            raise ContentException(
                "The provider can only be a single string or a tuple of strings, but got '%s'."
                % source
            )
        try:
            the_cmd = self.cmd % source
            return CommandOutputProvider(
                the_cmd,
                ctx,
                save_as=self.save_as,
                split=self.split,
                keep_rc=self.keep_rc,
                ds=self,
                timeout=self.timeout,
                inherit_env=self.inherit_env,
                override_env=self.override_env,
                signum=self.signum,
                cleaner=cleaner,
            )
        except NoFilterException as nfe:
            raise nfe
        except ContentException as ce:
            log.debug(ce)
        except Exception:
            log.debug(traceback.format_exc())
        raise ContentException("No results found for [%s]" % self.cmd)


class foreach_execute(object):
    """
    Execute a command for each element in provider. Provider is the output of
    a different datasource that returns a list of single elements or a list of
    tuples. The command should have %s substitution parameters equal to the
    number of elements in each tuple of the provider.

    Args:
        provider (list): a list of elements or tuples.
        cmd (str): a command with substitution parameters. Breaking
            apart a command string that might contain multiple commands
            separated by a pipe, getting them ready for subproc operations.
            IE. A command with filters applied
        context (ExecutionContext): the context under which the datasource
            should run.
        split (bool): whether the output of the command should be split into a
            list of lines
        keep_rc (bool): whether to return the error code returned by the
            process executing the command. If False, any return code other than
            zero with raise a CalledProcessError. If True, the return code and
            output are always returned.
        timeout (int): Number of seconds to wait for the command to complete.
            If the timeout is reached before the command returns, a
            CalledProcessError is raised. If None, timeout is infinite.
        inherit_env (list): The list of environment variables to inherit from the
            calling process when the command is invoked.
        override_env (dict): A dict of environment variables to override from the
            calling process when the command is invoked.

    Returns:
        function: A datasource that returns a list of outputs for each command
            created by substituting each element of provider into the cmd template.
    """

    def __init__(
        self,
        provider,
        cmd,
        context=HostContext,
        deps=None,
        split=True,
        keep_rc=False,
        timeout=None,
        inherit_env=None,
        override_env=None,
        signum=None,
        **kwargs
    ):
        deps = deps if deps is not None else []
        self.provider = provider
        self.cmd = cmd
        self.context = context
        self.split = split
        self.raw = not split
        self.keep_rc = keep_rc
        self.timeout = timeout
        self.inherit_env = inherit_env if inherit_env is not None else []
        self.override_env = override_env if override_env is not None else dict()
        self.signum = signum
        self.__name__ = self.__class__.__name__
        datasource(self.provider, self.context, *deps, multi_output=True, raw=self.raw, **kwargs)(
            self
        )

    def __call__(self, broker):
        result = []
        source = broker[self.provider]
        cleaner = broker.get('cleaner')
        ctx = broker[self.context]
        if isinstance(source, ContentProvider):
            source = source.content
        if not isinstance(source, (list, set)):
            source = [source]
        for e in source:
            try:
                the_cmd = self.cmd % e
                cop = CommandOutputProvider(
                    the_cmd,
                    ctx,
                    args=e,
                    split=self.split,
                    keep_rc=self.keep_rc,
                    ds=self,
                    timeout=self.timeout,
                    inherit_env=self.inherit_env,
                    override_env=self.override_env,
                    signum=self.signum,
                    cleaner=cleaner,
                )
                result.append(cop)
            except NoFilterException as nfe:
                raise nfe
            except ContentException as ce:
                log.debug(ce)
            except Exception:
                log.debug(traceback.format_exc())
        if result:
            return result
        raise ContentException("No results found for [%s]" % self.cmd)


class foreach_collect(object):
    """
    Subtitutes each element in provider into path and collects the files at the
    resulting paths.

    Args:
        provider (list): a list of elements or tuples.
        save_as (str or None): directory path to save the collected files as.
            - It should be a relative path and any starting '/' will be removed
            and an ending '/' will be added.
        path (str): a path template with substitution parameters.
        context (ExecutionContext): the context under which the datasource
            should run.
        kind (FileProvider): one of TextFileProvider or RawFileProvider

    Returns:
        function: A datasource that returns a list of file contents created by
            substituting each element of provider into the path template.
    """

    def __init__(
        self,
        provider,
        path,
        save_as=None,
        ignore=None,
        context=HostContext,
        deps=None,
        kind=TextFileProvider,
        **kwargs
    ):
        deps = deps if deps is not None else []
        self.provider = provider
        self.path = path
        self.save_as = os.path.join(save_as.lstrip("/"), '') if save_as else None
        self.ignore = ignore
        self.ignore_func = re.compile(ignore).search if ignore else lambda x: False
        self.context = context
        self.kind = kind
        self.raw = kind is RawFileProvider
        self.__name__ = self.__class__.__name__
        datasource(self.provider, self.context, *deps, multi_output=True, raw=self.raw, **kwargs)(
            self
        )

    def __call__(self, broker):
        result = []
        source = broker[self.provider]
        cleaner = broker.get('cleaner')
        ctx = _get_context(self.context, broker)
        root = ctx.root
        if isinstance(source, ContentProvider):
            source = source.content
        if not isinstance(source, (list, set)):
            source = [source]
        for e in source:
            pattern = ctx.locate_path(self.path % e)
            for p in glob(os.path.join(root, pattern.lstrip('/'))):
                if self.ignore_func(p) or os.path.isdir(p):
                    continue
                try:
                    result.append(
                        self.kind(
                            p[len(root) :],
                            root=root,
                            save_as=self.save_as,
                            ds=self,
                            ctx=ctx,
                            cleaner=cleaner,
                        )
                    )
                except NoFilterException as nfe:
                    raise nfe
                except Exception:
                    log.debug(traceback.format_exc())
        if result:
            return result
        raise ContentException("No results found for [%s]" % self.path)


class container_execute(foreach_execute):
    """
    Execute a command for each element in provider in container. Provider is
    the output of a different datasource that returns a list of tuples. In each
    tuple, the container engine provider ("podman" or "docker") and the
    container_id are two required elements, the rest elements if there are, are
    the arguments being passed to the command.

    Args:
        provider (list): a list of tuples, in each tuple, the container engine
            provider ("podman" or "docker") and the container_id are two
            required elements, the rest elements if there are, are the
            arguments being passed to the `cmd`.
        cmd (str): a command with substitution parameters. Breaking
            apart a command string that might contain multiple commands
            separated by a pipe, getting them ready for subproc operations.
            IE. A command with filters applied
        context (ExecutionContext): the context under which the datasource
            should run.
        split (bool): whether the output of the command should be split into a
            list of lines
        keep_rc (bool): whether to return the error code returned by the
            process executing the command. If False, any return code other than
            zero with raise a CalledProcessError. If True, the return code and
            output are always returned.
        timeout (int): Number of seconds to wait for the command to complete.
            If the timeout is reached before the command returns, a
            CalledProcessError is raised. If None, timeout is infinite.
        inherit_env (list): The list of environment variables to inherit from the
            calling process when the command is invoked.

    Returns:
        function: A datasource that returns a list of outputs for each command
            created by substituting each element of provider into the cmd template.
    """

    def __call__(self, broker):
        result = []
        source = broker[self.provider]
        cleaner = broker.get('cleaner')
        ctx = broker[self.context]
        if isinstance(source, ContentProvider):
            source = source.content
        if not isinstance(source, (list, set)):
            source = [source]
        for e in source:
            try:
                # e       = (image, <podman|docker>, container_id, <args>)
                image, engine, cid, args = e[0], e[1], e[2], e[3:]
                # handle command with args
                cmd = self.cmd % args if args else self.cmd
                # the_cmd = <podman|docker> exec container_id cmd
                the_cmd = "/usr/bin/%s exec %s %s" % (engine, cid, cmd)
                ccp = ContainerCommandProvider(
                    the_cmd,
                    ctx,
                    image=image,
                    args=e,
                    split=self.split,
                    keep_rc=self.keep_rc,
                    ds=self,
                    timeout=self.timeout,
                    inherit_env=self.inherit_env,
                    override_env=self.override_env,
                    signum=self.signum,
                    cleaner=cleaner,
                )
                result.append(ccp)
            except NoFilterException as nfe:
                raise nfe
            except Exception:
                log.debug(traceback.format_exc())
        if result:
            return result
        raise ContentException("No results found for [%s]" % self.cmd)


class container_collect(foreach_execute):
    """
    Collects the files at the resulting path in running containers.

    Args:
        provider (list): a list of tuples.
        path (str): the file path template with substitution parameters.  The
            path can also be passed via the provider when it's variable per
            cases, in that case, the `path` should be None.
        context (ExecutionContext): the context under which the datasource
            should run.
        keep_rc (bool): whether to return the error code returned by the
            process executing the command. If False, any return code other than
            zero with raise a CalledProcessError. If True, the return code and
            output are always returned.
        timeout (int): Number of seconds to wait for the command to complete.
            If the timeout is reached before the command returns, a
            CalledProcessError is raised. If None, timeout is infinite.

    Returns:
        function: A datasource that returns a list of file contents created by
            substituting each element of provider into the path template.
    """

    def __init__(
        self,
        provider,
        path=None,
        context=HostContext,
        deps=None,
        split=True,
        keep_rc=False,
        timeout=None,
        inherit_env=None,
        override_env=None,
        signum=None,
        **kwargs
    ):
        super(container_collect, self).__init__(
            provider,
            path,
            context,
            deps,
            split,
            keep_rc,
            timeout,
            inherit_env,
            override_env,
            signum,
            **kwargs
        )

    def __call__(self, broker):
        result = []
        source = broker[self.provider]
        cleaner = broker.get('cleaner')
        ctx = broker[self.context]
        if isinstance(source, ContentProvider):
            source = source.content
        if not isinstance(source, (list, set)):
            source = [source]
        for e in source:
            try:
                # e       = (image, <podman|docker>, container_id, <path>)
                image, e = e[0], e[1:]
                if self.cmd is None or self.cmd == '%s':
                    # path is provided by `provider`
                    e, path = e[:-1], e[-1]
                else:
                    # path is provided by self.cmd
                    e, path = e, self.cmd
                # e       = (<podman|docker>, container_id)
                # the_cmd = <podman|docker> exec container_id cat path
                the_cmd = ("/usr/bin/%s exec %s cat " % e) + path
                cfp = ContainerFileProvider(
                    the_cmd,
                    ctx,
                    image=image,
                    args=None,
                    split=self.split,
                    keep_rc=self.keep_rc,
                    ds=self,
                    timeout=self.timeout,
                    inherit_env=self.inherit_env,
                    override_env=self.override_env,
                    signum=self.signum,
                    cleaner=cleaner,
                )
                result.append(cfp)
            except NoFilterException as nfe:
                raise nfe
            except Exception:
                log.debug(traceback.format_exc())
        if result:
            return result
        raise ContentException("No results found for [%s]" % self.cmd)


class first_of(object):
    """Given a list of dependencies, returns the first of the list
    that exists in the broker. At least one must be present, or this
    component won't fire.
    """

    def __init__(self, deps):
        self.deps = deps
        self.raw = getattr(deps[0], 'raw', None)
        self.__name__ = self.__class__.__name__
        datasource(deps)(self)

    def __call__(self, broker):
        for c in self.deps:
            if c in broker:
                return broker[c]


class find(object):
    """
    Helper class for extracting specific lines from a datasource for direct
    consumption by a rule.

    .. code:: python

        service_starts = find(Specs.audit_log, "SERVICE_START")

        @rule(service_starts)
        def report(starts):
            return make_info("SERVICE_STARTS", num_starts=len(starts))

    Args:
        spec (datasource): some datasource, ideally filterable.
        pattern (string / list): a string or list of strings to match (no
            patterns supported)

    Returns:
        A dict where each key is a command, path, or spec name, and each value
        is a non-empty list of matching lines. Only paths with matching lines
        are included.

    Raises:
        SkipComponent: if no paths have matching lines.
    """

    def __init__(self, spec, pattern):
        if getattr(spec, "raw", False):
            name = dr.get_name(spec)
            raise ValueError("{}: Cannot filter raw files.".format(name))

        self.spec = spec
        self.pattern = pattern if isinstance(pattern, list) else [pattern]
        self.__name__ = self.__class__.__name__
        self.__module__ = self.__class__.__module__

        if getattr(spec, "filterable", False):
            filters._add_filter(spec, pattern)

        component(spec)(self)

    def __call__(self, ds):
        # /usr/bin/grep level filtering is applied behind .content or
        # .stream(), but we still need to ensure we get only what *this* find
        # instance wants. This can be inefficient on files where many lines
        # match.
        results = {}
        ds = ds if isinstance(ds, list) else [ds]
        for d in ds:
            if d.relative_path:
                origin = os.path.join("/", d.relative_path.lstrip("/"))
            elif d.cmd:
                origin = d.cmd
            else:
                origin = dr.get_name(self.spec)
            stream = d.content if d.loaded else d.stream()
            lines = []
            for line in stream:
                if any(p in line for p in self.pattern):
                    lines.append(line)
            if lines:
                results[origin] = lines
        if not results:
            raise SkipComponent()
        return dict(results)


@serializer(CommandOutputProvider)
def serialize_command_output(obj, root):
    rel = os.path.join("insights_commands", obj.relative_path)
    if obj.save_as:
        rel = os.path.join("insights_commands", obj.save_as)
        if obj.save_as.endswith("/"):
            rel = os.path.join(rel, os.path.basename(obj.relative_path))
    dst = os.path.join(root, rel)
    rc = obj.write(dst)
    return {
        "rc": rc,
        "cmd": obj.cmd,
        "args": obj.args,
        "save_as": bool(obj.save_as),
        "relative_path": rel,
    }


@deserializer(CommandOutputProvider)
def deserialize_command_output(_type, data, root, ctx, ds):
    rel = data["relative_path"]

    res = SerializedOutputProvider(rel, root=root, ctx=ctx, ds=ds)

    res.rc = data["rc"]
    res.cmd = data["cmd"]
    res.args = data["args"]
    return res


@serializer(TextFileProvider)
def serialize_text_file_provider(obj, root):
    rel = obj.relative_path
    if obj.save_as:
        rel = obj.save_as
        if obj.save_as.endswith("/"):
            rel = os.path.join(rel, os.path.basename(obj.relative_path))
    dst = os.path.join(root, rel)
    rc = obj.write(dst)
    return {
        "save_as": bool(obj.save_as),
        "relative_path": rel,
        "rc": rc,
    }


@deserializer(TextFileProvider)
def deserialize_text_provider(_type, data, root, ctx, ds):
    rel = data["relative_path"]
    res = SerializedOutputProvider(rel, root=root, ctx=ctx, ds=ds)
    res.rc = data["rc"]
    return res


@serializer(RawFileProvider)
def serialize_raw_file_provider(obj, root):
    rel = obj.relative_path
    if obj.save_as:
        rel = obj.save_as
        if obj.save_as.endswith("/"):
            rel = os.path.join(rel, os.path.basename(obj.relative_path))
    dst = os.path.join(root, rel)
    rc = obj.write(dst)
    return {
        "save_as": bool(obj.save_as),
        "relative_path": rel,
        "rc": rc,
    }


@deserializer(RawFileProvider)
def deserialize_raw_file_provider(_type, data, root, ctx, ds):
    rel = data["relative_path"]
    res = SerializedRawOutputProvider(rel, root=root, ctx=ctx, ds=ds)
    res.rc = data["rc"]
    return res


@serializer(DatasourceProvider)
def serialize_datasource_provider(obj, root):
    rel = obj.relative_path
    if obj.save_as:
        rel = obj.save_as
        if obj.save_as.endswith("/"):
            rel = os.path.join(rel, os.path.basename(obj.relative_path))
    dst = os.path.join(root, rel)
    obj.write(dst)
    return {"relative_path": rel, "save_as": obj.save_as}


@deserializer(DatasourceProvider)
def deserialize_datasource_provider(_type, data, root, ctx, ds):
    res = SerializedOutputProvider(data["relative_path"], root=root, ctx=ctx, ds=ds)
    return res


@serializer(MetadataProvider)
def serialize_metadata_provider(obj, root):
    # Built-in metadata files are put in the root instead of '/data'
    root = os.path.dirname(root) if os.path.basename(root) == 'data' else root
    rel = obj.relative_path
    if obj.save_as:
        rel = obj.save_as
        if obj.save_as.endswith("/"):
            rel = os.path.join(rel, os.path.basename(obj.relative_path))
    dst = os.path.join(root, obj.relative_path)
    obj.write(dst)
    return {"relative_path": rel, "save_as": obj.save_as}


@deserializer(MetadataProvider)
def deserialize_metadata_provider(_type, data, root, ctx, ds):
    # Built-in metadata files are put in the root instead of '/data'
    root = os.path.dirname(root) if os.path.basename(root) == 'data' else root
    res = SerializedOutputProvider(data["relative_path"], root=root, ctx=ctx, ds=ds)
    return res


@serializer(ContainerFileProvider)
def serialize_container_file_output(obj, root):
    rel = os.path.join("insights_containers", obj.relative_path)
    if obj.save_as:
        rel = os.path.join("insights_containers", obj.save_as)
        if obj.save_as.endswith("/"):
            rel = os.path.join(rel, os.path.basename(obj.relative_path))
    dst = os.path.join(root, rel)
    rc = obj.write(dst)
    return {
        "save_as": bool(obj.save_as),
        "relative_path": rel,
        "rc": rc,
        "image": obj.image,
        "engine": obj.engine,
        "container_id": obj.container_id,
    }


@deserializer(ContainerFileProvider)
def deserialize_container_file(_type, data, root, ctx, ds):
    rel = data["relative_path"]
    res = SerializedOutputProvider(rel, root=root, ctx=ctx, ds=ds)
    res.rc = data["rc"]
    res.image = data["image"]
    res.engine = data["engine"]
    res.container_id = data["container_id"]
    return res


@serializer(ContainerCommandProvider)
def serialize_container_command(obj, root):
    rel = os.path.join("insights_containers", obj.relative_path)
    if obj.save_as:
        rel = os.path.join("insights_containers", obj.save_as)
        if obj.save_as.endswith("/"):
            rel = os.path.join(rel, os.path.basename(obj.relative_path))
    dst = os.path.join(root, rel)
    rc = obj.write(dst)
    return {
        "rc": rc,
        "cmd": obj.cmd,
        "args": obj.args,
        "save_as": bool(obj.save_as),
        "relative_path": rel,
        "image": obj.image,
        "engine": obj.engine,
        "container_id": obj.container_id,
    }


@deserializer(ContainerCommandProvider)
def deserialize_container_command(_type, data, root, ctx, ds):
    rel = data["relative_path"]
    res = SerializedOutputProvider(rel, root=root, ctx=ctx, ds=ds)
    res.rc = data["rc"]
    res.cmd = data["cmd"]
    res.args = data["args"]
    res.image = data["image"]
    res.engine = data["engine"]
    res.container_id = data["container_id"]
    return res



"""
LDIF Configuration - file ``/etc/dirsrv/slapd-*/dse.ldif``
==========================================================
"""
from insights.core import Parser
from insights.core.exceptions import SkipComponent
from insights.core.plugins import parser
from insights.parsers import keyword_search
from insights.specs import Specs
from insights.util import deprecated


@parser(Specs.ldif_config)
class LDIFParser(Parser, list):
    """
    .. warning::
        This class is deprecated and will be removed from 3.5.0.
        Please use the :class:`insights.parsers.dse_ldif.DseLDIF` instead.

    Parse the content of the directory server configuration of the
    ``/etc/dirsrv/slapd-*/dse.ldif`` file.

    The file dse.ldif is in the LDIF format. LDIF contains multi-row records
    where each record is identified by a ``dn:`` line ("dn" as in "distinguished
    name") and the record's other lines are attributes. The value may be specified
    as UTF-8 text or as base64 encoded data, or a URI may be provided to the
    location of the attribute value.

    .. note::
        1. This parser unwraps the multiple 'aci:' lines to a single line.
        2. This parser only keeps the last value of a multiple keys and
           discrads the others before it.

    Sample output::

        dn:
        aci: (targetattr != "aci")(version 3.0; aci "rootdse anon read access"; allow(
         read,search,compare) userdn="ldap:///anyone";)
        aci: (target = "ldap:///cn=automember rebuild membership,cn=tasks,cn=config")(
         ,cn=permissions,cn=pbac,dc=idm";)
        createTimestamp: 20201026161200Z
        creatorsName: cn=server,cn=plugins,cn=config
        modifiersName: cn=Directory Manager
        modifyTimestamp: 20210608144722Z
        nsslapd-return-default-opattr: namingContexts
        nsslapd-return-default-opattr: supportedControl
        nsslapd-return-default-opattr: supportedExtension
        nsslapd-return-default-opattr: supportedLDAPVersion
        nsslapd-return-default-opattr: supportedSASLMechanisms
        nsslapd-return-default-opattr: vendorName
        nsslapd-return-default-opattr: vendorVersion
        objectClass: top

        dn: cn=changelog5,cn=config
        cn: changelog5
        createTimestamp: 20201026161228Z
        creatorsName: cn=Directory Manager
        modifiersName: cn=Directory Manager
        modifyTimestamp: 20201026161228Z
        nsslapd-changelogdir: /var/lib/dirsrv/slapd-IDM-NYPD-FINEST/cldb
        nsslapd-changelogmaxage: 7d
        objectClass: top
        objectClass: extensibleobject

    Returns:

        list: A list of dictionaries for each 'dn' attribute block of the ldif configuration.

    Examples:

        >>> ldif_config[0]['dn']
        ''
        >>> ldif_config[0]['aci']  # the 2 aci are connected into one
        '(targetattr != "aci")(version 3.0; aci "rootdse anon read access"; allow(read,search,compare) userdn="ldap:///anyone";)(target = "ldap:///cn=automember rebuild membership,cn=tasks,cn=config")(,cn=permissions,cn=pbac,dc=idm";)'
        >>> ldif_config[0]['nsslapd-return-default-opattr']  # only keep the last
        'vendorVersion'
        >>> ldif_config[1]['dn']
        'cn=changelog5,cn=config'
        >>> ldif_config[1]['modifiersName']
        'cn=Directory Manager'
        >>> ldif_config[1]['modifyTimestamp']
        '20201026161228Z'
    """

    def __init__(self, *args, **kwargs):
        deprecated(LDIFParser, "Please use the :class:`insights.parsers.dse_ldif.DseLDIF` instead.", "3.5.0")
        super(LDIFParser, self).__init__(*args, **kwargs)

    def parse_content(self, content):
        if not content:
            raise SkipComponent('The file is empty')

        attr_kval = {}
        for line in content:
            # lines beginning with # are ignored
            line = line.strip()
            if not line or line.startswith('#'):
                continue
            # line starts with 'dn' attribute
            elif line.startswith('dn:'):
                aci_flag = False
                attr_kval = dict(dn=line.split(':', 1)[1].strip())
                self.append(attr_kval)
            # line starts with 'aci' attribute
            elif line.startswith('aci:'):
                if 'aci' in attr_kval:
                    attr_kval['aci'] += line.split(':', 1)[1].strip()
                else:
                    attr_kval['aci'] = line.split(':', 1)[1].strip()
                aci_flag = True
            # line is a muti-line value with the 'aci' attribute
            elif aci_flag and ': ' not in line:
                attr_kval['aci'] += line
            # line is a non 'aci' attribute or file-backed value attribute
            elif not line.startswith('aci:') and ': ' in line:
                aci_flag = False
                key, val = [i.strip() for i in line.split(':', 1)]
                attr_kval[key] = val

    def search(self, **kwargs):
        """
        Get the list for the 'dn' attribute block by searching the ldif configuration.
        This uses the :py:func:`insights.parsers.keyword_search` function for searching,
        see its documentation for usage details. If no search parameters are given or does
        match the search, then nothing will be returned.

        Returns:
            list: A list of dictionaries for each 'dn' attribute block of the ldif configuration that match the given
            search criteria.

        Examples:
            >>> ldif_config.search(dn__contains='cn=config')[0] == ldif_config[1]
            True
            >>> ldif_config.search(dn='cn=sasl,cn=config') == []
            True
            >>> ldif_config.search(cn='changelog5')[0] == ldif_config[1]
            True
        """
        return keyword_search(self, row_keys_change=True, **kwargs)



"""
RHUI release commands or files
==============================

RHUISetRelease - command ``rhui_set_release``
---------------------------------------------
RHUIReleaseVer - file ``/etc/yum/vars/releasever`` or ``/etc/yum/vars/releasever``
----------------------------------------------------------------------------------
"""
import os

from insights.core import Parser, CommandParser
from insights.core.exceptions import SkipComponent
from insights.core.plugins import parser
from insights.specs import Specs
from insights.util import deprecated


@parser(Specs.rhui_releasever)
class RHUIReleaseVer(Parser):
    """
    Class for parsing the file `/etc/yum/vars/releasever` or `/etc/dnf/vars/releasever`.
    It will output the rhel minor release when a minor release is set,
    or emtpy when it isn't set. This file may not exist when there is no minor release
    is set.

    Sample input data::

        8.6

    Attributes:
        set (str): the set release.
        major (int): the major version of the set release.
        minor (int): the minor version of the set release.

    Raises:
       SkipComponent: When the content is not in expected format.

    Examples:
        >>> type(rhui_releasever)
        <class 'insights.parsers.rhui_release.RHUIReleaseVer'>
        >>> rhui_releasever.set
        '8.6'
        >>> rhui_releasever.major
        8
        >>> rhui_releasever.minor
        6
    """

    def parse_content(self, content):
        self.set = self.major = self.minor = None
        if len(content) == 0:
            # Release not set
            return
        if len(content) != 1:
            raise SkipComponent("Unexpected content: {0}".format(os.linesep.join(content)))
        rhel_version = content[0].strip()
        line_splits = rhel_version.split('.')
        if len(line_splits) == 2 and line_splits[0].isdigit() and line_splits[-1].isdigit():
            self.set = rhel_version
            self.major = int(line_splits[0])
            self.minor = int(line_splits[-1])
        elif rhel_version and rhel_version[0].isdigit():
            self.set = rhel_version
            self.major = int(rhel_version[0])
            # leave self.minor as None
        else:
            self.set = rhel_version


@parser(Specs.rhui_set_release)
class RHUISetRelease(CommandParser):
    """
    .. warning::
        This class is deprecated and will be removed from 3.6.0.
        Please use the :class:`insights.parsers.rhui_release.RHUIReleaseVer` instead.

    Class for parsing the output of `rhui_set_release` command.
    It will output the rhel minor release when a minor release is set,
    or emtpy when it isn't set.

    Typical output of the command is::

        8.6

    Attributes:
        set (str): the set release.
        major (int): the major version of the set release.
        minor (int): the minor version of the set release.

    Examples:
        >>> type(rhui_rel)
        <class 'insights.parsers.rhui_release.RHUISetRelease'>
        >>> rhui_rel.set
        '8.6'
        >>> rhui_rel.major
        8
        >>> rhui_rel.minor
        6
    """

    def __init__(self, context, extra_bad_lines=None):
        deprecated(RHUISetRelease, "Please use the :class:`insights.parsers.rhui_release.RHUIReleaseVer` instead.", "3.6.0")
        super(RHUISetRelease, self).__init__(context, extra_bad_lines)

    def parse_content(self, content):
        self.set = self.major = self.minor = None
        if len(content) == 0:
            # Release not set
            return
        if len(content) == 1:
            rhel_version = content[0].strip()
            line_splits = rhel_version.split('.')
            if len(line_splits) == 2 and line_splits[0].isdigit() and line_splits[-1].isdigit():
                self.set = rhel_version
                self.major = int(line_splits[0])
                self.minor = int(line_splits[-1])
                return
        raise SkipComponent("Unexpected content: {0}".format(os.linesep.join(content)))



import doctest

from collections import defaultdict

from insights.core import filters
from insights.parsers import ls_usr_bin
from insights.parsers.ls_usr_bin import LsUsrBin
from insights.specs import Specs
from insights.tests import context_wrap

LS_USR_BIN = """
total 41472
lrwxrwxrwx.  1 0  0        7 Oct 22  2019 python -> python2
-rwxr-xr-x.  1 0  0     2558 Apr 10  2019 python-argcomplete-check-easy-install-script
-rwxr-xr-x.  1 0  0      318 Apr 10  2019 python-argcomplete-tcsh
lrwxrwxrwx.  1 0  0       14 Oct 22  2019 python-config -> python2-config
lrwxrwxrwx.  1 0  0        9 Oct 22  2019 python2 -> python2.7
"""


def teardown_function(func):
    filters._CACHE = {}
    filters.FILTERS = defaultdict(set)


def test_ls_usr_bin():
    ls_usr_bin = LsUsrBin(context_wrap(LS_USR_BIN, path='insights_commands/ls_-ln_.usr.bin'))
    assert ls_usr_bin.files_of('/usr/bin') == ['python', 'python-argcomplete-check-easy-install-script', 'python-argcomplete-tcsh', 'python-config', 'python2']
    python = ls_usr_bin.dir_entry('/usr/bin', 'python')
    assert python is not None
    assert python == {'date': 'Oct 22  2019',
        'dir': '/usr/bin',
        'group': '0',
        'link': 'python2',
        'links': 1,
        'name': 'python',
        'owner': '0',
        'perms': 'rwxrwxrwx.',
        'raw_entry': 'lrwxrwxrwx.  1 0  0        7 Oct 22  2019 python -> python2',
        'size': 7,
        'type': 'l'}


def test_ls_usr_bin_doc_examples():
    env = {
        'Specs': Specs,
        'add_filter': filters.add_filter,
        'LsUsrBin': LsUsrBin,
        'ls_usr_bin': LsUsrBin(context_wrap(LS_USR_BIN, path='insights_commands/ls_-ln_.usr.bin')),
    }
    failed, total = doctest.testmod(ls_usr_bin, globs=env)
    assert failed == 0



import pytest

from mock.mock import Mock

from insights.core.exceptions import SkipComponent
from insights.core.spec_factory import DatasourceProvider
from insights.specs.datasources.lpstat import LocalSpecs, lpstat_protocol_printers_info, lpstat_queued_jobs_count


LPSTAT_V = """
device for test_printer1: ipp://cups.test.com/printers/test_printer1
device for test_printer2: ipp://cups.test.com/printers/test_printer2
device for test_printer3: socket://192.168.1.5:9100
device for test_printer4: usb://smth
device for test_printer5: ///dev/null
""".strip()

LPSTAT_V_NOT_GOOD = """
lpstat: Transport endpoint is not connected
""".strip()

LPSTAT_V_RESULT = """
device for test_printer1: ipp
device for test_printer2: ipp
device for test_printer3: socket
device for test_printer4: usb
device for test_printer5: ///dev/null
""".strip()

RELATIVE_PATH = 'insights_commands/lpstat_-v'

LPSTAT_O = """
Cups-PDF-1802           root          265443328   Tue 05 Sep 2023 02:21:19 PM CST
Cups-PDF-1803           root          265443328   Tue 05 Sep 2023 02:21:21 PM CST
Cups-PDF-1804           root          265443328   Tue 05 Sep 2023 02:21:22 PM CST
""".strip()

LPSTAT_O_EMPTY = """
""".strip()

LPSTAT_O_INVALID = """
lpstat: Bad file descriptor
""".strip()

LPSTAT_O_RESULT = """
3
""".strip()

LPSTAT_O_EMPTY_RESULT = """
0
""".strip()

LPSTAT_O_PATH = 'insights_commands/lpstat_-o_jobs_count'


def test_lpstat_datasource():
    lpstat_data = Mock()
    lpstat_data.content = LPSTAT_V.splitlines()
    broker = {LocalSpecs.lpstat_v: lpstat_data}
    result = lpstat_protocol_printers_info(broker)
    assert result is not None
    assert isinstance(result, DatasourceProvider)
    expected = DatasourceProvider(content=LPSTAT_V_RESULT, relative_path=RELATIVE_PATH)
    assert result.content == expected.content
    assert result.relative_path == expected.relative_path


def test_lpstat_datasource_NG_output():
    lpstat_data = Mock()
    lpstat_data.content = LPSTAT_V_NOT_GOOD.splitlines()
    broker = {LocalSpecs.lpstat_v: lpstat_data}
    with pytest.raises(SkipComponent) as e:
        lpstat_protocol_printers_info(broker)
    assert 'SkipComponent' in str(e)


def test_lpstat_o():
    lpstat_o = Mock()
    lpstat_o.content = LPSTAT_O.splitlines()
    broker = {LocalSpecs.lpstat_o: lpstat_o}
    result = lpstat_queued_jobs_count(broker)
    assert isinstance(result, DatasourceProvider)
    expected = DatasourceProvider(content=LPSTAT_O_RESULT, relative_path=LPSTAT_O_PATH)
    assert result.content == expected.content
    assert result.relative_path == expected.relative_path


def test_lpstat_o_empty():
    lpstat_o = Mock()
    lpstat_o.content = LPSTAT_O_EMPTY.splitlines()
    broker = {LocalSpecs.lpstat_o: lpstat_o}
    with pytest.raises(SkipComponent) as e:
        lpstat_queued_jobs_count(broker)
    assert 'SkipComponent' in str(e)


def test_lpstat_o_invalid():
    lpstat_o = Mock()
    lpstat_o.content = LPSTAT_O_INVALID.splitlines()
    broker = {LocalSpecs.lpstat_o: lpstat_o}
    with pytest.raises(SkipComponent) as e:
        lpstat_queued_jobs_count(broker)
    assert 'SkipComponent' in str(e)



from insights.parsers.sysconfig import IfCFGStaticRoute
from insights.tests import context_wrap
import pytest

CONTEXT_PATH_DEVICE_1 = "etc/sysconfig/network-scripts/route-test-net"

STATIC_ROUTE_1 = """
ADDRESS0=10.65.223.0
NETMASK0=255.255.254.0
GATEWAY0=10.65.223.1
""".strip()

CONTEXT_PATH_DEVICE_2 = "etc/sysconfig/network-scripts/route-test-net-11"

STATIC_ROUTE_2 = """
ADDRESS0=
NETMASK0=255.255.254.0
GATEWAY0=10.65.223.1
""".strip()

CONTEXT_PATH_DEVICE_3 = "etc/sysconfig/network-scripts/ifcfg-eth0"


def test_static_route_connection_1():
    context = context_wrap(STATIC_ROUTE_1, path=CONTEXT_PATH_DEVICE_1)
    r = IfCFGStaticRoute(context)
    assert r.static_route_name == 'test-net'
    assert r.data == {'ADDRESS0': '10.65.223.0', 'NETMASK0': '255.255.254.0', 'GATEWAY0': '10.65.223.1'}


def test_static_route_connection_2():
    context = context_wrap(STATIC_ROUTE_1, CONTEXT_PATH_DEVICE_2)
    r = IfCFGStaticRoute(context)
    assert r.static_route_name == 'test-net-11'
    assert r.data == {'ADDRESS0': '10.65.223.0', 'NETMASK0': '255.255.254.0', 'GATEWAY0': '10.65.223.1'}


def test_missing_index():
    with pytest.raises(IndexError):
        context = context_wrap(STATIC_ROUTE_2, CONTEXT_PATH_DEVICE_3)
        r = IfCFGStaticRoute(context)
        assert r.static_route_name == 'eth0'



import doctest
import pytest

from insights.core.exceptions import ParseException
from insights.parsers import lsscsi
from insights.tests import context_wrap


LSSCSI_1 = """
[1:0:0:0]    storage IET      Controller       0001  -
[1:0:0:1]    cd/dvd  QEMU     QEMU DVD-ROM     2.5+  /dev/sr0
[1:0:0:2]    disk    IET      VIRTUAL-DISK     0001  /dev/sdb
[3:0:5:0]    tape    HP       C5713A           H910  /dev/st0
"""

LSSCSI_2 = """
[1:0:0:1]    cd/dvd  QEMU     QEMU DVD-ROM     2.5+  /dev/sr0
"""

LSSCSI_3 = """
[1:0:0:1]    cd/dvd  QEMU     QEMU  DVD-ROM     2.5+  /dev/sr0
"""

LSSCSI_4 = """
[1:0:0:1]    cd/dvd  QEMU     QEMU DVD-ROM     2.5+  /dev/sr0
[1:0:0:2]    cd/dvd  QEMU     QEMU DVD-ROM     2.5+  /dev/sr1
[1:0:0:3]    disk    IET      VIRTUAL-DISK     0001  /dev/sdb
[3:0:5:0]    tape    HP       C5713A           H910  /dev/st0
"""

LSSCSI_5 = """
[1:0:0:2]    disk    IET      VIRTUAL-DISK     0001
[3:0:5:0]    tape    HP       C5713A           H910  /dev/st0
"""


def test_lsscsi():
    scsi = lsscsi.LsSCSI(context_wrap(LSSCSI_1))
    assert len(scsi.data) == 4
    assert scsi[0] == {'Model': 'Controller', 'Vendor': 'IET',
                         'HCTL': '[1:0:0:0]', 'Peripheral-Type': 'storage',
                         'Primary-Device-Node': '-', 'Revision': '0001'}
    assert scsi[1]['Peripheral-Type'] == 'cd/dvd'
    assert ['-', '/dev/sr0', '/dev/sdb', '/dev/st0'] == scsi.device_nodes

    assert scsi[1]['Vendor'] == 'QEMU'
    assert ['IET', 'QEMU', 'IET', 'HP'] == scsi.device_vendors

    scsi = lsscsi.LsSCSI(context_wrap(LSSCSI_2))
    assert len(scsi.data) == 1
    assert scsi[0] == {'Model': 'QEMU DVD-ROM', 'Vendor': 'QEMU',
                       'HCTL': '[1:0:0:1]', 'Peripheral-Type': 'cd/dvd',
                       'Primary-Device-Node': '/dev/sr0', 'Revision': '2.5+'}

    scsi = lsscsi.LsSCSI(context_wrap(LSSCSI_3))
    assert len(scsi.data) == 1
    assert scsi[0]['Model'] == 'QEMU  DVD-ROM'

    scsi = lsscsi.LsSCSI(context_wrap(LSSCSI_4))
    assert len(scsi.data) == 4
    assert len(scsi[0]) == 6


def test_bad_lsscsi():
    with pytest.raises(ParseException) as e_info:
        lsscsi.LsSCSI(context_wrap(""))
    assert "Empty content of command output" in str(e_info.value)

    with pytest.raises(ParseException) as e_info:
        lsscsi.LsSCSI(context_wrap(LSSCSI_5))
    assert "Invalid format of content, unparsable" in str(e_info.value)


def test_lsscsi_documentation():
    failed_count, tests = doctest.testmod(
        lsscsi,
        globs={'lsscsi': lsscsi.LsSCSI(context_wrap(LSSCSI_1))}
    )
    assert failed_count == 0



import glob
import os
import json
import pytest
import tempfile

from collections import defaultdict
from mock.mock import patch, Mock

from insights import collect
from insights.client.archive import InsightsArchive
from insights.client.config import InsightsConfig
from insights.core import Parser, dr
from insights.core import filters
from insights.core.context import HostContext
from insights.core.exceptions import ContentException
from insights.core.filters import add_filter
from insights.core.plugins import datasource
from insights.core.spec_cleaner import Cleaner
from insights.core.spec_factory import (
        DatasourceProvider, RegistryPoint, SpecSet, command_with_args,
        foreach_collect, foreach_execute,
        glob_file, simple_command, simple_file, first_file, first_of)
from insights.specs.default import _make_rpm_formatter

here = os.path.abspath(os.path.dirname(__file__))

DATA = """
Some test data
"""
MAX_GLOBS = 1001

this_file = os.path.abspath(__file__).rstrip("c")

test_empty_file = '/tmp/_insights_test.empty_file'
test_empty_after_filter = '/tmp/_insights_test.empty_after_filter'
test_empty_after_redact = '/tmp/_insights_test.empty_after_redact'

specs_manifest = """
---
version: 0

client:
    context:
        class: insights.core.context.HostContext
        args:
            timeout: 5

    # commands and files to ignore
    blacklist:
        files: []
        commands: []
        patterns: ['TeSt_']
        keywords: []

    persist:
        - name: insights.tests.test_specs.Specs
          enabled: true

    run_strategy:
        name: serial
        args:
            max_workers: null

plugins:
    default_component_enabled: false

    packages:
        - insights.tests.test_specs

    configs:
        - name: insights.core.spec_factory
          enabled: true
        - name: insights.tests.test_specs.Specs
          enabled: true
        - name: insights.tests.test_specs.Stuff
          enabled: true
        - name: insights.tests.test_specs.dostuff
          enabled: true
""".strip()


# fixture to set up max glob file test
@pytest.fixture
def max_globs():
    for f in range(MAX_GLOBS):
        tmpfile = tempfile.NamedTemporaryFile(prefix='tmp_', suffix='_glob', delete=False)
        tmpfile.close()
        with open(tmpfile.name, 'w') as fd:
            fd.write(DATA)
    yield tempfile.gettempdir()
    for fle in glob.glob1(tempfile.gettempdir(), "tmp_*_glob"):
        os.remove(tempfile.gettempdir() + "/" + fle)


class Specs(SpecSet):
    many_glob = RegistryPoint(multi_output=True, no_redact=True)
    many_glob_filter = RegistryPoint(multi_output=True, no_redact=True, filterable=True)
    many_foreach_exe = RegistryPoint(multi_output=True)
    many_foreach_exe_filter = RegistryPoint(multi_output=True, filterable=True)
    many_foreach_clc = RegistryPoint(multi_output=True)
    many_foreach_clc_filter = RegistryPoint(multi_output=True, filterable=True)
    smpl_cmd = RegistryPoint()
    smpl_cmd_w_filter = RegistryPoint(filterable=True)
    smpl_file = RegistryPoint()
    smpl_file_w_filter = RegistryPoint(filterable=True)
    first_file_spec_w_filter = RegistryPoint(filterable=True)
    first_of_spec_w_filter = RegistryPoint(filterable=True)
    no_such_cmd = RegistryPoint()
    no_such_file = RegistryPoint()
    empty_orig = RegistryPoint()
    empty_after_filter = RegistryPoint(filterable=True)
    empty_after_redact = RegistryPoint()
    cmd_w_args_filter = RegistryPoint(filterable=True)


class Stuff(Specs):
    many_glob = glob_file(here + "/test_a*.py")
    many_glob_filter = glob_file(here + "/test_b*.py")

    @datasource(HostContext)
    def files0(broker):
        """ Return a list of directories from the spec filter """
        return [here + "/__init__.py", here + "/integration.py"]

    @datasource(HostContext)
    def files1(broker):
        """ Return a list of directories from the spec filter """
        return [here + "/test_test.py", here + "/test_taglang.py"]

    @datasource(HostContext)
    def files2(broker):
        """ Return a list of directories from the spec filter """
        return ' '.join([here + "/test_test.py", here + "/test_taglang.py"])

    many_foreach_exe = foreach_execute(files0, 'ls -t %s')
    many_foreach_exe_filter = foreach_execute(files1, 'ls -l %s')
    many_foreach_clc = foreach_collect(files0, "%s", no_redact=True)
    many_foreach_clc_filter = foreach_collect(files1, "%s", no_redact=True)
    smpl_cmd = simple_command("/usr/bin/uptime")
    smpl_cmd_w_filter = simple_command("echo -n ' hello 1'")
    smpl_file = simple_file(here + "/helpers.py")
    smpl_file_w_filter = simple_file(here + "/mock_web_server.py")
    first_file_spec_w_filter = first_file([here + "/spec_tests.py", "/etc/os-release"])
    first_of_spec_w_filter = first_of(
        [
            simple_command("echo -n ' hello 1'"),
            simple_file(this_file),
        ]
    )
    no_such_cmd = simple_command("/usr/bin/no_such_cmd args")
    no_such_file = simple_file("/no/such/_file")

    empty_orig = simple_file(test_empty_file)
    empty_after_filter = simple_file(test_empty_after_filter)
    empty_after_redact = simple_file(test_empty_after_redact)
    cmd_w_args_filter = command_with_args('ls -lt %s', files2)


class stage(dr.ComponentType):
    def invoke(self, broker):
        return self.component(broker)


@stage(
    Stuff.many_glob,
    Stuff.many_glob_filter,
    Stuff.many_foreach_exe,
    Stuff.many_foreach_exe_filter,
    Stuff.many_foreach_clc,
    Stuff.many_foreach_clc_filter,
    Stuff.smpl_cmd,
    Stuff.smpl_cmd_w_filter,
    Stuff.smpl_file,
    Stuff.smpl_file_w_filter,
    Stuff.first_file_spec_w_filter,
    Stuff.first_of_spec_w_filter,
    Stuff.cmd_w_args_filter,
    optional=[Stuff.no_such_cmd,
              Stuff.empty_orig,
              Stuff.empty_after_filter,
              Stuff.empty_after_redact,
              Stuff.no_such_file, ])
def dostuff(broker):
    assert Stuff.many_glob in broker
    assert Stuff.many_glob_filter in broker
    assert Stuff.many_foreach_exe in broker
    assert Stuff.many_foreach_exe_filter in broker
    assert Stuff.many_foreach_clc in broker
    assert Stuff.many_foreach_clc_filter in broker
    assert Stuff.smpl_cmd in broker
    assert Stuff.smpl_cmd_w_filter in broker
    assert Stuff.smpl_file in broker
    assert Stuff.smpl_file_w_filter in broker
    assert Stuff.first_file_spec_w_filter in broker
    assert Stuff.first_of_spec_w_filter in broker
    assert Stuff.cmd_w_args_filter in broker

    assert Stuff.empty_orig not in broker
    assert Stuff.empty_after_filter not in broker
    assert Stuff.empty_after_redact not in broker
    assert Stuff.no_such_cmd not in broker
    assert Stuff.no_such_file not in broker


# File content
with open(here + "/helpers.py") as f:
    smpl_file_content = f.read().splitlines()
with open(here + "/mock_web_server.py") as f:
    smpl_file_w_filter_content = [l for l in f.read().splitlines() if "def get" in l]
with open(here + "/spec_tests.py") as f:
    first_file_w_filter_content = [l for l in f.read().splitlines() if any(i in l for i in ["def report_", "rhel"])]

#
# TEST
#


def setup_function(func):
    if func == test_specs_collect:
        # empty relevant files
        open(test_empty_file, 'a').close()
        with open(test_empty_after_filter, 'w') as t:
            t.write('no-filter')
        with open(test_empty_after_redact, 'w') as t:
            t.write('TeSt_')


def teardown_function(func):
    filters._CACHE = {}
    filters.FILTERS = defaultdict(set)

    if func == test_specs_collect:
        os.remove(test_empty_file)
        os.remove(test_empty_after_filter)
        os.remove(test_empty_after_redact)


def test_spec_factory():
    add_filter(Stuff.many_glob_filter, " ")
    add_filter(Stuff.many_foreach_exe_filter, " ")
    add_filter(Stuff.many_foreach_clc_filter, " ")
    add_filter(Stuff.smpl_cmd_w_filter, " hello ")
    add_filter(Stuff.smpl_file_w_filter, "def get")
    add_filter(Stuff.first_file_spec_w_filter, ["def report_", "rhel"])
    add_filter(Stuff.first_of_spec_w_filter, ["def test", " hello "])
    add_filter(Stuff.empty_after_filter, " hello ")
    add_filter(Stuff.cmd_w_args_filter, [" ", ":"])
    broker = dr.Broker()
    broker[HostContext] = HostContext()
    broker['cleaner'] = Cleaner(None, None)
    broker = dr.run(dr.get_dependency_graph(dostuff), broker)

    assert dostuff in broker, broker.tracebacks
    assert broker[Stuff.smpl_file].content == smpl_file_content
    assert not any(l.endswith("\n") for l in broker[Stuff.smpl_file].content)
    # "filter" works when loading
    assert "hello" in broker[Stuff.smpl_cmd_w_filter].content[0]
    assert len(broker[Stuff.smpl_cmd_w_filter].content) == 1
    assert broker[Stuff.smpl_file_w_filter].content == smpl_file_w_filter_content
    assert broker[Stuff.first_file_spec_w_filter].content == first_file_w_filter_content
    assert len(broker[Stuff.first_of_spec_w_filter].content) == 1
    assert len(broker.exceptions) == 5
    for exp in broker.exceptions:
        if "no_such" in str(exp):
            assert "args" not in str(exp)
        else:
            assert "/no/such" not in str(exp)
    assert len(broker.tracebacks) == 5


def test_line_terminators():
    add_filter(Stuff.many_glob_filter, " ")
    add_filter(Stuff.many_foreach_exe_filter, " ")
    add_filter(Stuff.many_foreach_clc_filter, " ")
    add_filter(Stuff.smpl_cmd_w_filter, " hello ")
    add_filter(Stuff.smpl_file_w_filter, "def get")
    add_filter(Stuff.first_file_spec_w_filter, ["def report_", "rhel"])
    add_filter(Stuff.first_of_spec_w_filter, ["def test", " hello "])
    add_filter(Stuff.empty_after_filter, " hello ")
    add_filter(Stuff.cmd_w_args_filter, [" ", ":"])
    broker = dr.Broker()
    broker[HostContext] = HostContext()
    broker['cleaner'] = Cleaner(None, None)
    broker = dr.run(dr.get_dependency_graph(dostuff), broker)

    content = broker[Stuff.smpl_file_w_filter].content
    assert not any(l.endswith("\n") for l in content)
    # "filter" works only when writing
    # assert all("def get" in l for l in content), content
    assert "hello" in broker[Stuff.smpl_cmd_w_filter].content[0]
    assert len(broker[Stuff.smpl_cmd_w_filter].content) == 1
    assert len(broker[Stuff.smpl_file_w_filter].content) == len(smpl_file_w_filter_content)
    assert len(broker[Stuff.first_file_spec_w_filter].content) == len(first_file_w_filter_content)
    assert len(broker[Stuff.first_of_spec_w_filter].content) == 1


def test_glob_max(max_globs):
    too_many = glob_file(max_globs + "/tmp_*_glob")
    broker = dr.Broker()
    broker[HostContext] = HostContext()
    broker['cleaner'] = Cleaner(None, None)
    with pytest.raises(ContentException):
        too_many(broker)


def test_datasource_provider():
    data = "blahblah\nblahblah2"

    class MyParser(Parser):
        def parse_content(self, content):
            self.content = content

    ds = DatasourceProvider(data, relative_path="things")
    p = MyParser(ds)
    assert p.content == data.splitlines()
    assert list(ds.stream()) == data.splitlines()


def test_exp_no_filters():
    broker = dr.Broker()
    broker[HostContext] = HostContext()
    broker['cleaner'] = Cleaner(None, None)
    broker = dr.run(dr.get_dependency_graph(dostuff), broker)
    assert dostuff not in broker
    exception_cnt = 0
    for spec, info in broker.exceptions.items():
        # warning of lacking filters
        if any("due to" in str(msg) for msg in info):
            if "smpl_file_w_filter" in str(spec):
                exception_cnt += 1
            elif "smpl_cmd_w_filter" in str(spec):
                exception_cnt += 10
            elif "first_file_spec_w_filter" in str(spec):
                exception_cnt += 100
            elif "first_of_spec_w_filter" in str(spec):
                exception_cnt += 1000
            elif "many_glob_filter" in str(spec):
                exception_cnt += 10000
            elif "many_foreach_exe_filter" in str(spec):
                exception_cnt += 100000
            elif "many_foreach_clc_filter" in str(spec):
                exception_cnt += 1000000
            elif "cmd_w_args_filter" in str(spec):
                exception_cnt += 10000000
    assert exception_cnt == 11111111


@pytest.mark.parametrize("obfuscate", [True, False])
@patch('insights.core.spec_cleaner.Cleaner.generate_report', Mock())
def test_specs_collect(obfuscate):
    add_filter(Stuff.many_glob_filter, " ")
    add_filter(Stuff.many_foreach_exe_filter, " ")
    add_filter(Stuff.many_foreach_clc_filter, " ")
    add_filter(Stuff.smpl_cmd_w_filter, " hello ")
    add_filter(Stuff.smpl_file_w_filter, "def get")
    add_filter(Stuff.first_file_spec_w_filter, [" hello ", "Test"])
    add_filter(Stuff.first_of_spec_w_filter, ["def test", " hello "])
    add_filter(Stuff.empty_after_filter, " hello ")
    add_filter(Stuff.cmd_w_args_filter, [" ", ":"])
    # Preparation
    manifest = collect.load_manifest(specs_manifest)
    for pkg in manifest.get("plugins", {}).get("packages", []):
        dr.load_components(pkg, exclude=None)
    # For verifying convenience, test obfuscate=False only
    conf = InsightsConfig(
            obfuscate=obfuscate, obfuscate_hostname=obfuscate,
            manifest=manifest)
    arch = InsightsArchive(conf)
    arch.create_archive_dir()
    output_path, errors = collect.collect(
            tmp_path=arch.tmp_dir,
            archive_name=arch.archive_name,
            client_config=conf)
    meta_data_root = os.path.join(output_path, 'meta_data')
    data_root = os.path.join(output_path, 'data')

    assert not errors
    count = 0
    for spec in Specs.__dict__:
        if not spec.startswith(('__', 'context_handlers', 'registry')):
            file_name = "insights.tests.test_specs.Specs.{0}.json".format(spec)
            meta_data = os.path.join(meta_data_root, file_name)
            with open(meta_data, 'r') as fp:
                count += 1
                mdata = json.load(fp)
                results = mdata.get('results')
                if 'no_such' in spec:
                    assert results is None
                    continue
                if "empty" in spec:
                    assert results is None
                    assert ": Empty " in mdata.get('errors')[0]
                    continue
                if not isinstance(results, list):
                    results = [results]
                for result in results:
                    if not result:
                        continue
                    rel = result['object']['relative_path']
                    org_content = new_content = None
                    with open(os.path.join(data_root, rel), 'r') as fp:
                        new_content = fp.readlines()
                    if "insights_commands" not in rel:
                        # check files only
                        with open("/" + rel, 'r') as fp:
                            org_content = fp.readlines()
                        if "_filter" not in spec:
                            if org_content:
                                if obfuscate:
                                    assert len(org_content) == len(new_content)
                                    continue
                                    # Test below only for obfuscate=False
                                assert org_content[:-1] == new_content[:-1]
                                # Special: no '\n' in last line of collected data
                                assert org_content[-1].strip() == new_content[-1].strip()
                            else:
                                # Both Empty
                                assert not new_content
                        else:
                            # "filter" works in archive result files
                            assert len(org_content) > len(new_content)
    assert count == 18  # Number of Specs

    arch.delete_archive_dir()

    # Reset Test ENV
    dr.COMPONENTS = defaultdict(lambda: defaultdict(set))
    dr.TYPE_OBSERVERS = defaultdict(set)
    dr.ENABLED = defaultdict(lambda: True)


def test_specs_default_module_utils():
    rpm_formatter = _make_rpm_formatter([
        '"name":"%{NAME}"',
        '"version":"%{VERSION}"'
    ])
    assert ',"version":' in rpm_formatter



import doctest
import pytest

from insights.core.exceptions import SkipComponent
from insights.parsers import subscription_manager
from insights.parsers.subscription_manager import SubscriptionManagerID, SubscriptionManagerFacts, SubscriptionManagerStatus
from insights.tests import context_wrap

FACTS_NORMAL_1 = """
aws_instance_id: 567890567890
network.ipv6_address: ::1
uname.sysname: Linux
uname.version: #1 SMP PREEMPT Fri Sep 2 16:07:40 EDT 2022
virt.host_type: rhev, kvm
virt.is_guest: True
""".strip()

FACTS_with_AB_LINES = """
You are attempting to use a locale that is not installed.
aws_instance_id: 567890567890
network.ipv6_address: ::1
uname.sysname: Linux
uname.version: #1 SMP PREEMPT Fri Sep 2 16:07:40 EDT 2022
virt.host_type: rhev, kvm
virt.is_guest: True
""".strip()

ID_NORMAL_1 = """
system identity: 6655c27c-f561-4c99-a23f-f53e5a1ef311
name: rhel7.localdomain
org name: 1234567
org ID: 1234567
""".strip()

ID_with_AB_LINES = """
You are attempting to use a locale that is not installed.
system identity: 6655c27c-f561-4c99-a23f-f53e5a1ef311
name: rhel7.localdomain
org name: 1234567
org ID: 1234567
""".strip()

INPUT_NG_1 = """
XYC
Release not set
""".strip()

INPUT_NG_2 = ""


SUBSCRIPTION_MANAGER_STATUS = """
+-------------------------------------------+
   System Status Details
+-------------------------------------------+
Overall Status: Disabled
Content Access Mode is set to Simple Content Access. This host has access to content, regardless of subscription status.

System Purpose Status: Disabled
""".strip()

SUBSCRIPTION_MANAGER_STATUS_EMPTY = """

""".strip()

SUBSCRIPTION_MANAGER_STATUS_PRODUCT_KEY = """
+-------------------------------------------+
   System Status Details
+-------------------------------------------+
Overall Status: Insufficient

Red Hat Enterprise Linux for Virtual Datacenters, Standard:
- Guest has not been reported on any host and is using a temporary unmapped guest subscription. For more information, please see https://access.redhat.com/solutions/XXXX

System Purpose Status: Matched
""".strip()


def test_subman_facts():
    ret = SubscriptionManagerFacts(context_wrap(FACTS_NORMAL_1))
    for line in FACTS_NORMAL_1.splitlines():
        key, value = line.split(': ', 1)
        assert ret[key] == value

    ret = SubscriptionManagerFacts(context_wrap(FACTS_with_AB_LINES))
    for line in FACTS_NORMAL_1.splitlines():
        if ': ' not in line:
            continue
        key, value = line.split(': ', 1)
        assert ret[key] == value


def test_subman_id():
    ret = SubscriptionManagerID(context_wrap(ID_NORMAL_1))
    for line in ID_NORMAL_1.splitlines():
        key, value = line.split(': ', 1)
        assert ret[key] == value
    assert ret.identity == '6655c27c-f561-4c99-a23f-f53e5a1ef311'

    ret = SubscriptionManagerID(context_wrap(ID_with_AB_LINES))
    for line in ID_with_AB_LINES.splitlines():
        if ': ' not in line:
            continue
        key, value = line.split(': ', 1)
        assert ret[key] == value
    assert ret.identity == '6655c27c-f561-4c99-a23f-f53e5a1ef311'


def test_subman_status():
    ret = SubscriptionManagerStatus(context_wrap(SUBSCRIPTION_MANAGER_STATUS))
    assert ret['Overall Status'] == 'Disabled'
    assert ret['Content Access Mode'] == 'Simple Content Access'
    assert ret['System Purpose Status'] == 'Disabled'

    ret = SubscriptionManagerStatus(context_wrap(SUBSCRIPTION_MANAGER_STATUS_PRODUCT_KEY))
    assert ret['Overall Status'] == 'Insufficient'
    assert ret['Red Hat Enterprise Linux for Virtual Datacenters'] == 'Standard'
    assert ret['System Purpose Status'] == 'Matched'
    assert ret.unparsed_lines == [SUBSCRIPTION_MANAGER_STATUS_PRODUCT_KEY.splitlines()[-3]]


def test_subman_facts_ng():
    with pytest.raises(SkipComponent):
        SubscriptionManagerFacts(context_wrap(INPUT_NG_1))

    with pytest.raises(SkipComponent):
        SubscriptionManagerFacts(context_wrap(INPUT_NG_2))


def test_subman_id_ng():
    with pytest.raises(SkipComponent):
        SubscriptionManagerID(context_wrap(INPUT_NG_1))

    with pytest.raises(SkipComponent):
        SubscriptionManagerID(context_wrap(INPUT_NG_2))


def test_subman_status_ng():
    with pytest.raises(SkipComponent):
        SubscriptionManagerStatus(context_wrap(SUBSCRIPTION_MANAGER_STATUS_EMPTY))


def test_doc_examples():
    env = {
            'rhsm_facts': SubscriptionManagerFacts(context_wrap(FACTS_with_AB_LINES)),
            'subman_id': SubscriptionManagerID(context_wrap(ID_with_AB_LINES)),
            'subman_status': SubscriptionManagerStatus(context_wrap(SUBSCRIPTION_MANAGER_STATUS)),
          }
    failed, total = doctest.testmod(subscription_manager, globs=env)
    assert failed == 0



"""
Mlx4Port - file ``/sys/bus/pci/devices/*/mlx4_port[0-9]``
=========================================================

This module provides processing for the contents of each file
matching the glob spec ``/sys/bus/pci/devices/*/mlx4_port[0-9]``.

Sample contents of this file looks like::

    ib

or::

    eth

Attributes:
    name (str): The mlx4 port name.
    contents (list): List of string values representing each line in
        the file.

Examples:
    >>> type(mlx4_port)
    <class 'insights.parsers.mlx4_port.Mlx4Port'>
    >>> mlx4_port.name
    'mlx4_port1'
    >>> mlx4_port.contents
    ['ib']
"""

from .. import parser, Parser
from insights.specs import Specs


@parser(Specs.mlx4_port)
class Mlx4Port(Parser):
    """
    Parse the contents of the mlx4_port file
    """
    def parse_content(self, content):
        self.name = self.file_name
        self.contents = [line.strip() for line in content]



"""
LS Parsers - command ``ls``
===========================

Parsers provided in this module includes:

LSla - command ``ls -la <dirs>``
--------------------------------

LSlaFiltered - command ``ls -la <dirs> | grep -F <keywords>``
-------------------------------------------------------------

LSlan - command ``ls -lan <dirs>``
----------------------------------

LSlanFiltered - command ``ls -lan <dirs> | grep -F <keywords>``
---------------------------------------------------------------

LSlanL - command ``ls -lanL <dirs>``
------------------------------------

LSlanR - command ``ls -lanR <dirs>``
------------------------------------

LSlanRL - command ``ls -lanRL <dirs>``
--------------------------------------

LSlaRZ - command ``ls -lanRZ <dirs>``
-------------------------------------

LSlaZ - command ``ls -lanZ <dirs>``
-----------------------------------
"""
from insights.core import ls_parser, Parser
from insights.core.filters import add_filter
from insights.core.plugins import parser
from insights.specs import Specs
from insights.util.file_permissions import FilePermissions

# Required basic filters for `LS` specs that the content needs to be filtered
add_filter(Specs.ls_la_filtered, ['total '])
add_filter(Specs.ls_lan_filtered, ['total '])

# Required directories to collect for `LSlanR` specs, to:
# 1. keep compatible with sosreport archives
# 2. Ensure that the related applications to work properly:
#    - archive data extraction  (cee-data-engineering)
#    - insights-facts  (cee-data-engineering)
add_filter(Specs.ls_lanR_dirs, ['/boot', '/dev', '/sys/firmware'])


class FileListing(Parser, dict):
    """
    Reads a series of concatenated directory listings and turns them into
    a dictionary of entities by name.  Stores all the information for
    each directory entry for every entry that can be parsed, containing::

        - type (one of [bcdlps-])
        - permission string including ACL character
        - number of links
        - owner and group (as given in the listing)
        - size, or major and minor number for block and character devices
        - date (in the format given in the listing)
        - name
        - name of linked file, if a symlink

    In addition, the raw line is always stored, even if the line doesn't look
    like a directory entry.

    Also provides a number of other conveniences, such as::

        - lists of regular and special files and subdirectory names for each
          directory, in the order found in the listing
        - total blocks allocated to all the entities in this directory

    Parses the SELinux information if present in the listing.
    SELinux directory listings contain::

        - the type of file
        - the permissions block
        - the owner and group as given in the directory listing
        - the SELinux user, role, type and MLS
        - the name, and link destination if it's a symlink

    .. note::
        The :class:`FileListing` Parser parses the content collected by
        diffirent ``ls_*`` specs. The ``ls_*`` specs collect the corresponding
        ``ls`` command output according to the filters defined by the relevant
        ``ls_*_dirs`` specs.  For the ``ls_*_dirs`` specs, only absolute
        directory path is acceptable, path to file or relative path is not
        acceptable.  For details, see the following example.

    Sample output::

        /boot:
        total 187380
        dr-xr-xr-x.  3 0 0   4096 Mar  4 16:19 .
        dr-xr-xr-x. 19 0 0   4096 Jul 14 09:10 ..
        -rw-r--r--.  1 0 0 123891 Aug 25  2015 config-3.10.0-229.14.1.el7.x86_64

        /etc/sysconfig:
        total 96
        drwxr-xr-x.  7 0 0 4096 Jul  6 23:41 .
        drwxr-xr-x. 77 0 0 8192 Jul 13 03:55 ..
        drwxr-xr-x.  2 0 0   41 Jul  6 23:32 cbq
        drwxr-xr-x.  2 0 0    6 Sep 16  2015 console
        -rw-------.  1 0 0 1390 Mar  4  2014 ebtables-config
        -rw-r--r--.  1 0 0   72 Sep 15  2015 firewalld
        lrwxrwxrwx.  1 0 0   17 Jul  6 23:32 grub -> /etc/default/grub

    Examples:
        >>> from insights.core.filters import add_filter
        >>> from insights.specs import Specs
        >>> add_filter(Specs.ls_lan_dirs, ['/boot', '/etc/sysconfig'])
        >>> type(ls_lan)
        <class 'insights.parsers.ls.FileListing'>
        >>> "/etc" in ls_lan
        False
        >>> "/etc/sysconfig" in ls_lan
        True
        >>> len(ls_lan.files_of('/etc/sysconfig'))
        3
        >>> ls_lan.files_of("/etc/sysconfig")
        ['ebtables-config', 'firewalld', 'grub']
        >>> ls_lan.dirs_of("/etc/sysconfig")
        ['.', '..', 'cbq', 'console']
        >>> ls_lan.specials_of("/etc/sysconfig")
        []
        >>> ls_lan.total_of("/etc/sysconfig")
        96
        >>> ls_lan.dir_entry('/etc/sysconfig', 'grub') == {'group': '0', 'name': 'grub', 'links': 1, 'perms': 'rwxrwxrwx.', 'raw_entry': 'lrwxrwxrwx.  1 0 0   17 Jul  6 23:32 grub -> /etc/default/grub', 'owner': '0', 'link': '/etc/default/grub', 'date': 'Jul  6 23:32', 'type': 'l', 'dir': '/etc/sysconfig', 'size': 17}
        True
        >>> sorted(ls_lan.listing_of("/etc/sysconfig").keys()) == sorted(['console', 'grub', '..', 'firewalld', '.', 'cbq', 'ebtables-config'])
        True
        >>> sorted(ls_lan.listing_of("/etc/sysconfig")['console'].keys()) == sorted(['group', 'name', 'links', 'perms', 'raw_entry', 'owner', 'date', 'type', 'dir', 'size'])
        True
        >>> ls_lan.listing_of("/etc/sysconfig")['console']['type']
        'd'
        >>> ls_lan.listing_of("/etc/sysconfig")['console']['perms']
        'rwxr-xr-x.'
        >>> ls_lan.dir_contains("/etc/sysconfig", "console")
        True
        >>> ls_lan.dir_entry("/etc/sysconfig", "console") == {'group': '0', 'name': 'console', 'links': 2, 'perms': 'rwxr-xr-x.', 'raw_entry': 'drwxr-xr-x.  2 0 0    6 Sep 16  2015 console', 'owner': '0', 'date': 'Sep 16  2015', 'type': 'd', 'dir': '/etc/sysconfig', 'size': 6}
        True
        >>> ls_lan.dir_entry("/etc/sysconfig", "grub")['type']
        'l'
        >>> ls_lan.dir_entry("/etc/sysconfig", "grub")['link']
        '/etc/default/grub'
        >>> "/boot" in ls_lan
        True
        >>> ls_lan.files_of('/boot')
        ['config-3.10.0-229.14.1.el7.x86_64']
        >>> fp = ls_lan.permissions_of('/boot', 'config-3.10.0-229.14.1.el7.x86_64')
        >>> fp.owner
        '0'
        >>> fp.group
        '0'
        >>> fp.perms_owner
        'rw-'
    """
    __root_path = None
    """
    The root path of the dir when there is only one list target.  It only works
    for the following specs/parsers that are compatible for sos-archives.
    - :class:`insights.parsers.ls_boot.LsBoot`
    - :class:`insights.parsers.ls_dev.LsDev`
    - :class:`insights.parsers.ls_sys_firmware.LsSysFirmware`

    None by default, for the new ```ls_*``` datasource specs
    """

    def parse_content(self, content):
        """
        Called automatically to process the directory listing(s) contained in
        the content.
        """
        self.update(ls_parser.parse(content, self.__root_path))

    def files_of(self, directory):
        """
        The list of non-special files (i.e. not block or character files)
        in the given directory.
        """
        if directory in self:
            return self[directory]['files']
        return []

    def dirs_of(self, directory):
        """
        The list of subdirectories in the given directory.
        """
        if directory in self:
            return self[directory]['dirs']
        return []

    def specials_of(self, directory):
        """
        The list of block and character special files in the given directory.
        """
        if directory in self:
            return self[directory]['specials']
        return []

    def total_of(self, directory):
        """
        The total blocks of storage consumed by entries in this directory.
        """
        if directory in self:
            return self[directory]['total']
        return 0

    def listing_of(self, directory):
        """
        The listing of this directory, in a dictionary by entry name.  All
        entries contain the original line as is in the 'raw_entry' key.
        Entries that can be parsed then have fields as described in the class
        description above.
        """
        if directory in self:
            return self[directory]['entries']
        return []

    def dir_contains(self, directory, name):
        """
        Does this directory contain this entry name?
        """
        if directory in self:
            return name in self[directory]['entries']
        return False

    def dir_entry(self, directory, name):
        """
        The parsed data for the given entry name in the given directory.
        """
        if directory in self:
            return self[directory]['entries'][name]
        return {}

    def path_entry(self, path):
        """
        The parsed data given a path, which is separated into its directory
        and entry name.
        """
        if path[0] != '/':
            return None
        path_parts = path.split('/')
        # Note that here the first element will be '' because it's before the
        # first separator.  That's OK, the join puts it back together.
        directory = '/'.join(path_parts[:-1])
        if directory not in self:
            return None
        name = path_parts[-1]
        if name not in self[directory]['entries']:
            return None
        return self[directory]['entries'][name]

    def permissions_of(self, directory, target):
        """
        Returns a FilePermissions object, if found.

        Parameters:
            directory(string): Full path without trailing slash where to
                search.
            target (string): Name of the directory or file to get
                FilePermissions for.

        Returns:
            FilePermissions: If found or None if not found.
        """
        if directory in self:
            d = self[directory]['entries']
            if target in d:
                return FilePermissions(d[target]['raw_entry'])


@parser(Specs.ls_la)
class LSla(FileListing):
    """
    Parses output of ``ls -la <dirs>`` command.
    See :py:class:`FileListing` for more information.
    """
    pass


@parser(Specs.ls_la_filtered)
class LSlaFiltered(FileListing):
    """
    Parses output of ``ls -la <dirs> | grep -F <keywords>`` command.
    See :py:class:`FileListing` for more information.
    """
    pass


@parser(Specs.ls_lan)
class LSlan(FileListing):
    """
    Parses output of ``ls -lan <dirs>`` command.
    See :py:class:`FileListing` for more information.
    """
    pass


@parser(Specs.ls_lan_filtered)
class LSlanFiltered(FileListing):
    """
    Parses output of ``ls -lan <dirs> | grep -F <keywords>`` command.
    See :py:class:`FileListing` for more information.
    """
    pass


@parser(Specs.ls_lanL)
class LSlanL(FileListing):
    """
    Parses output of ``ls -lanR <dirs>`` command.
    See :py:class:`FileListing` for more information.
    """
    pass


@parser(Specs.ls_lanR)
class LSlanR(FileListing):
    """
    Parses output of ``ls -lanR <dirs>`` command.
    See :py:class:`FileListing` for more information.
    """
    pass


@parser(Specs.ls_lanRL)
class LSlanRL(FileListing):
    """
    Parses output of ``ls -lanRL <dirs>`` command.
    See :py:class:`FileListing` for more information.
    """
    pass


@parser(Specs.ls_laRZ)
class LSlaRZ(FileListing):
    """
    Parses output of ``ls -laRZ <dirs>`` command.
    See :py:class:`FileListing` for more information.
    """
    pass


@parser(Specs.ls_laZ)
class LSlaZ(FileListing):
    """
    Parses output of ``ls -laZ <dirs>`` command.
    See :py:class:`FileListing` for more information.
    """
    pass



"""
Parsers for VSFTPD configuration
================================

This module contains two parsers:

VsftpdPamConf - file ``/etc/pam.d/vsftpd``
------------------------------------------

VsftpdConf - file ``/etc/vsftpd.conf``
--------------------------------------

"""

from insights import Parser, LegacyItemAccess, parser
from insights.core import ContainerParser
from insights.parsers.pam import PamDConf
from insights.parsers import split_kv_pairs
from insights.specs import Specs


@parser(Specs.vsftpd)
class VsftpdPamConf(PamDConf):
    """
    Parsing for the `/etc/pam.d/vsftpd` PAM configuration.

    See the :py:class:`insights.parsers.pam.PamDConf` class documentation for
    more information on how this class is used.

    Sample PAM configuration for vsftpd::

        #%PAM-1.0
        session    optional     pam_keyinit.so    force revoke
        auth       required     pam_listfile.so item=user sense=deny file=/etc/vsftpd/ftpusers onerr=succeed
        auth       required     pam_shells.so
        auth       include      password-auth
        account    include      password-auth
        session    required     pam_loginuid.so
        session    include      password-auth

    Examples:
        >>> vsftpd_pam_conf[0].interface
        'session'
        >>> vsftpd_pam_conf[0].control_flags
        [ControlFlag(flag='optional', value=None)]
        >>> vsftpd_pam_conf[0].control_flags[0].flag
        'optional'
        >>> vsftpd_pam_conf[0].module_name
        'pam_keyinit.so'
        >>> vsftpd_pam_conf[0].module_args
        'force revoke'
    """
    pass


@parser(Specs.vsftpd_conf)
class VsftpdConf(Parser, LegacyItemAccess):
    """
    Parsing for `/etc/vsftpd/vsftpd.conf`.  Key=value pairs are stored in a
    dictionary, made available directly through the object itself thanks to
    the :py:class:`insights.core.LegacyItemAccess` mixin.

    Reference:
        https://access.redhat.com/documentation/en-US/Red_Hat_Enterprise_Linux/3/html/Reference_Guide/s1-ftp-vsftpd-conf.html

    Sample content::

        # No anonymous login
        anonymous_enable=NO
        # Let local users login
        local_enable=YES

        # Write permissions
        write_enable=YES

    Examples:
        >>> 'anonymous_enable' in vsftpd_conf
        True
        >>> 'chmod_enable' in vsftpd_conf
        False
        >>> vsftpd_conf['anonymous_enable']
        'NO'
    """
    def parse_content(self, content):
        self.data = split_kv_pairs(content, ordered=True)


@parser(Specs.container_vsftpd_conf)
class ContainerVsftpdConf(ContainerParser, VsftpdConf):
    """
    Parsing for `/etc/vsftpd/vsftpd.conf` from the containers.  Key=value pairs are stored in a
    dictionary, made available directly through the object itself thanks to
    the :py:class:`insights.core.LegacyItemAccess` mixin.

    Reference:
        https://access.redhat.com/documentation/en-US/Red_Hat_Enterprise_Linux/3/html/Reference_Guide/s1-ftp-vsftpd-conf.html

    Sample content::

        # No anonymous login
        anonymous_enable=NO
        # Let local users login
        local_enable=YES

        # Write permissions
        write_enable=YES

    Examples:
        >>> type(container_vsftpd_conf)
        <class 'insights.parsers.vsftpd.ContainerVsftpdConf'>
        >>> container_vsftpd_conf.container_id
        '2869b4e2541c'
        >>> container_vsftpd_conf.image
        'registry.access.redhat.com/ubi8/nginx-120'
        >>> 'anonymous_enable' in container_vsftpd_conf
        True
        >>> 'chmod_enable' in container_vsftpd_conf
        False
        >>> container_vsftpd_conf['anonymous_enable']
        'NO'
    """
    pass



import doctest

from insights.parsers import image_builder_facts
from insights.parsers.image_builder_facts import ImageBuilderFacts
from insights.tests import context_wrap
from insights.tests.parsers import skip_component_check


IB_FACTS = """
{
  "image-builder.insights.compliance-policy-id": "61812cce-c884-4d52-bf05-15b41370db23",
  "image-builder.insights.compliance-profile-id": "xccdf_org.ssgproject.content_profile_cis",
  "image-builder.osbuild-composer.api-type": "cloudapi-v2"
}
""".strip()


def test_image_builder_facts():
    result = ImageBuilderFacts(context_wrap(IB_FACTS))

    assert result.data == {
        "image-builder.insights.compliance-policy-id": "61812cce-c884-4d52-bf05-15b41370db23",
        "image-builder.insights.compliance-profile-id": "xccdf_org.ssgproject.content_profile_cis",
        "image-builder.osbuild-composer.api-type": "cloudapi-v2"
    }


def test_image_builder_facts_empty():
    assert 'Empty output.' in skip_component_check(ImageBuilderFacts)


def test_image_builder_facts_doc_examples():
    env = {
        'ImageBuilderFacts': ImageBuilderFacts,
        'image_builder_facts': ImageBuilderFacts(context_wrap(IB_FACTS)),
    }
    failed, total = doctest.testmod(image_builder_facts, globs=env)
    assert failed == 0



"""
Alternatives - command ``/usr/bin/alternatives`` output
=======================================================
"""

from insights import parser, CommandParser
from insights.core.exceptions import ParseException
from insights.specs import Specs


class AlternativesOutput(CommandParser):
    """
    Read the output of ``/usr/sbin/alternatives --display *program*`` and
    convert into information about the given program's alternatives.

    Typical input is::

        java - status is auto.
         link currently points to /usr/lib/jvm/java-1.8.0-openjdk-1.8.0.111-1.b15.el7_2.x86_64/jre/bin/java
        /usr/lib/jvm/java-1.7.0-openjdk-1.7.0.111-2.6.7.2.el7_2.x86_64/jre/bin/java - priority 1700111
        /usr/lib/jvm/java-1.8.0-openjdk-1.8.0.111-1.b15.el7_2.x86_64/jre/bin/java - priority 1800111
        /usr/lib/jvm/jre-1.6.0-ibm.x86_64/bin/java - priority 16091
         slave ControlPanel: /usr/lib/jvm/jre-1.6.0-ibm.x86_64/bin/ControlPanel
         slave keytool: /usr/lib/jvm/jre-1.6.0-ibm.x86_64/bin/keytool
         slave policytool: /usr/lib/jvm/jre-1.6.0-ibm.x86_64/bin/policytool
         slave rmid: /usr/lib/jvm/jre-1.6.0-ibm.x86_64/bin/rmid
        Current 'best' version is /usr/lib/jvm/jre-1.6.0-ibm.x86_64/bin/java.

    Lines are interpreted this way:

    * Program lines are of the form '*name* - status is *status*', and start
      the information for a program.  Lines before this are ignored.
    * The current link to this program is found on lines starting with 'link
      currently points to'.
    * Lines starting with '/' and with ' - priority ' in them record an
      alternative path and its priority.
    * Lines starting with 'slave *program*: *path*' are recorded against the
      alternative path.
    * Lines starting with "Current 'best' version is" indicate the default
      choice of an 'auto' status alternative.

    The output of ``alternatives --display *program*`` can only ever list one
    program, so as long as one 'status is' line is found (as described above),
    the content of the object displays that program.

    Attributes:
        program (str): The name of the program found in the 'status is' line.
            This attribute is set to ``None`` if a status line is not found.
        status (str): The status of the program, or ``None`` if not found.
        link (str): The link to this program, or ``None`` if the 'link
            currently points to`` line is not found.
        best (str): The 'best choice' path that ``alternatives`` would use, or
            ``None`` if the 'best choice' line is not found.
        paths (dict): The alternative paths for this program.  Each path is a
            dictionary containing the following keys:

              * ``path``: the actual path of this alternative for the program
              * ``priority``: the priority, as an integer (e.g. 1700111)
              * ``slave``: a dictionary of programs dependent on this alternative -
                the key is the program name (e.g. 'ControlPanel') and the value is
                the path to that program for this alternative path.

    Examples:
        >>> java_alt.program
        'java'
        >>> java_alt.link
        '/usr/lib/jvm/jre-1.6.0-ibm.x86_64/bin/java'
        >>> len(java_alt.paths)
        2
        >>> java_alt.paths[0]['path']
        '/usr/lib/jvm/jre-1.6.0-ibm.x86_64/bin/java'
        >>> java_alt.paths[0]['priority']
        16091
        >>> java_alt.paths[0]['slave']['ControlPanel']
        '/usr/lib/jvm/jre-1.6.0-ibm.x86_64/bin/ControlPanel'
    """

    def parse_content(self, content):
        """
        Parse the output of the ``alternatives`` command.
        """
        self.program = None
        self.status = None
        self.link = None
        self.best = None
        self.paths = []
        current_path = None
        # Set up instance variable

        for line in content:
            words = line.split(None)
            if ' - status is' in line:
                # alternatives only displays one program, so finding
                # this line again is an error.
                if self.program:
                    raise ParseException(
                        "Program line for {newprog} found in output for {oldprog}".format(
                            newprog=words[0], oldprog=self.program
                        )
                    )
                # Set up new program data
                self.program = words[0]
                self.status = words[4][:-1]  # remove trailing .
                self.alternatives = []
                current_path = {}
            elif not self.program:
                # Lines before 'status is' line are ignored
                continue
            elif line.startswith(' link currently points to ') and len(words) == 5:
                # line: ' link currently points to /usr/lib/jvm/java-1.8.0-openjdk-1.8.0.111-1.b15.el7_2.x86_64/jre/bin/java'
                self.link = words[4]
            elif ' - priority ' in line and len(words) == 4 and words[3].isdigit():
                # line: /usr/lib/jvm/jre-1.6.0-ibm.x86_64/bin/java - priority 16091
                # New path - save current path if set
                self.paths.append({
                    'path': words[0],
                    'priority': int(words[3]),
                    'slave': {},
                })
                current_path = self.paths[-1]
            elif line.startswith(' slave ') and len(words) == 3 and current_path:
                # line: ' slave ControlPanel: /usr/lib/jvm/jre-1.6.0-ibm.x86_64/bin/ControlPanel'
                current_path['slave'][words[1][:-1]] = words[2]  # remove final : from program
            elif line.startswith("Current `best' version is ") and len(words) == 5:
                # line: 'Current `best' version is /usr/lib/jvm/jre-1.6.0-ibm.x86_64/bin/java.'
                self.best = words[4][:-1]  # strip trailing . from path


@parser(Specs.display_java)
class JavaAlternatives(AlternativesOutput):
    """
    Class to read the ``/usr/sbin/alternatives --display java`` output.

    Uses the ``AlternativesOutput`` base class to get information about the
    alternatives for ``java`` available and which one is currently in use.

    Examples:
        >>> java_alt.program
        'java'
        >>> java_alt.link
        '/usr/lib/jvm/jre-1.6.0-ibm.x86_64/bin/java'
        >>> len(java_alt.paths)
        2
        >>> java_alt.paths[0]['path']
        '/usr/lib/jvm/jre-1.6.0-ibm.x86_64/bin/java'
        >>> java_alt.paths[0]['priority']
        16091
        >>> java_alt.paths[0]['slave']['ControlPanel']
        '/usr/lib/jvm/jre-1.6.0-ibm.x86_64/bin/ControlPanel'
    """
    pass


@parser(Specs.alternatives_display_python)
class PythonAlternatives(AlternativesOutput):
    """
    Class to read the ``/usr/sbin/alternatives --display python`` output.

    Uses the ``AlternativesOutput`` base class to get information about the
    alternatives for ``best`` available and which one is currently in use.

    Examples:
        >>> python_alt.program
        'python'
        >>> python_alt.link
        '/usr/bin/python3'
        >>> len(python_alt.paths)
        2
        >>> python_alt.paths[0]['path']
        '/usr/libexec/no-python'
        >>> python_alt.paths[0]['priority']
        404
    """
    pass



import pytest
import sys

from collections import defaultdict

from insights import parser
from insights.combiners.hostname import Hostname
from insights.core import filters
from insights.core.spec_factory import RegistryPoint, SpecSet, simple_file
from insights.parsers.ps import PsAux, PsAuxcww
from insights.specs import Specs
from insights.specs.default import DefaultSpecs


class MySpecs(SpecSet):
    no_filters = RegistryPoint(filterable=False)
    has_filters = RegistryPoint(filterable=True)


class LocalSpecs(MySpecs):
    no_filters = simple_file('no_filters')
    has_filters = simple_file('has_filters')


@parser(MySpecs.has_filters)
class MySpecsHasFilters(object):
    pass


@parser(LocalSpecs.has_filters)
class LocalSpecsHasFilters(object):
    pass


@parser(MySpecs.no_filters)
class LocalSpecsNoFilters(object):
    pass

#
# TEST
#


def setup_function(func):
    if func is test_get_filter:
        filters.add_filter(Specs.ps_aux, "COMMAND")

    if func is test_get_filter_registry_point:
        filters.add_filter(Specs.ps_aux, "COMMAND")
        filters.add_filter(DefaultSpecs.ps_aux, "MEM")

    if func is test_filter_dumps_loads:
        filters.add_filter(Specs.ps_aux, ["PID", "COMMAND", "TEST"])


def teardown_function(func):
    filters._CACHE = {}
    filters.FILTERS = defaultdict(set)


@pytest.mark.skipif(sys.version_info < (2, 7), reason='Playbook verifier code uses oyaml library which is incompatable with this test')
def test_filter_dumps_loads():
    r = filters.dumps()
    assert r is not None

    filters.loads(r)

    assert Specs.ps_aux in filters.FILTERS
    assert filters.FILTERS[Specs.ps_aux] == set(["PID", "COMMAND", "TEST"])

    r2 = filters.dumps()
    assert r2 == r  # 'filters' are in the same order in every dumps()


def test_get_filter():
    f = filters.get_filters(Specs.ps_aux)
    assert "COMMAND" in f

    f = filters.get_filters(DefaultSpecs.ps_aux)
    assert "COMMAND" in f


def test_get_filter_registry_point():
    s = set(["COMMAND", "MEM"])
    f = filters.get_filters(DefaultSpecs.ps_aux)
    assert f & s == s

    f = filters.get_filters(Specs.ps_aux)
    assert "COMMAND" in f
    assert "MEM" not in f


# Local Parser
def test_add_filter_to_MySpecsHasFilters():
    """
    "filters" added to MySpecs.x will also add to LocalSpecs.x
    """
    filter_string = "bash"

    # Local Parser depends on MySpecs (Specs)
    filters.add_filter(MySpecsHasFilters, filter_string)

    myspecs_filters = filters.get_filters(MySpecs.has_filters)
    assert filter_string in myspecs_filters
    assert filter_string in filters.FILTERS[MySpecs.has_filters]

    localspecs_filters = filters.get_filters(LocalSpecs.has_filters)
    assert filter_string in localspecs_filters
    # but there is no key in FILTERS for the LocalSpecs.x
    assert filter_string not in filters.FILTERS[LocalSpecs.has_filters]


def test_add_filter_to_LocalSpecsHasFilters():
    """
    "filters" added to LocalSpecs.x will NOT add to Specs.x
    """
    filter_string = "bash"
    filters.add_filter(LocalSpecsHasFilters, filter_string)

    myspecs_filters = filters.get_filters(MySpecs.has_filters)
    assert filter_string not in myspecs_filters
    assert filter_string not in filters.FILTERS[MySpecs.has_filters]

    localspecs_filters = filters.get_filters(LocalSpecs.has_filters)
    assert filter_string in localspecs_filters
    assert filter_string in filters.FILTERS[LocalSpecs.has_filters]


# General Parser
def test_add_filter_to_PsAux():
    """
    "filters" added to Specs.x will add to DefaultSpecs.x
    """
    filter_string = "bash"
    filters.add_filter(PsAux, filter_string)

    spec_filters = filters.get_filters(Specs.ps_aux)
    assert filter_string in spec_filters
    assert filter_string in filters.FILTERS[Specs.ps_aux]

    default_spec_filters = filters.get_filters(DefaultSpecs.ps_aux)
    assert filter_string in default_spec_filters  # get_filters() works
    # but there is no key in FILTERS for the LocalSpecs.x
    assert filter_string not in filters.FILTERS[DefaultSpecs.ps_aux]  # empty in FILTERS

    parser_filters = filters.get_filters(PsAux)
    assert not parser_filters


def test_add_filter_to_parser_patterns_list():
    filters_list = ["bash", "systemd", "Network"]
    filters.add_filter(PsAux, filters_list)

    spec_filters = filters.get_filters(Specs.ps_aux)
    assert all(f in spec_filters for f in filters_list)

    parser_filters = filters.get_filters(PsAux)
    assert not parser_filters


def test_add_filter_exception_spec_not_filterable():
    with pytest.raises(Exception):
        filters.add_filter(Specs.ps_auxcww, "bash")


def test_add_filter_exception_parser_non_filterable():
    with pytest.raises(Exception):
        filters.add_filter(PsAuxcww, 'bash')

    with pytest.raises(Exception):
        filters.add_filter(LocalSpecsNoFilters, 'bash')


def test_add_filter_exception_combiner_non_filterable():
    with pytest.raises(Exception):
        filters.add_filter(Hostname, "bash")


def test_add_filter_exception_raw():
    with pytest.raises(Exception):
        filters.add_filter(Specs.metadata_json, "[]")


def test_add_filter_exception_empty():
    with pytest.raises(Exception):
        filters.add_filter(Specs.ps_aux, "")


def test_get_filters():
    _filter = 'A filter'
    filters.add_filter(MySpecs.has_filters, _filter)

    ret_has = filters.get_filters(MySpecs.has_filters)
    assert _filter in ret_has
    ret_no = filters.get_filters(MySpecs.no_filters)
    assert len(ret_no) == 0



import doctest
import pytest

from insights.core.exceptions import ParseException
from insights.parsers import sys_block
from insights.tests import context_wrap

STABLE_WRITES = """
1
""".strip()

STABLE_WRITES_INVALID = """
invalid
""".strip()

STABLE_WRITES_EMPTY = """
""".strip()

FILE_PATH_SDA = "/sys/block/sda/queue/scheduler"


def test_stable_writes():
    res = sys_block.StableWrites(context_wrap(STABLE_WRITES, FILE_PATH_SDA))
    assert res.stable_writes == 1
    assert res.device == 'sda'


def test_invalid_stable_writes():
    with pytest.raises(ParseException) as e:
        sys_block.StableWrites(context_wrap(STABLE_WRITES_INVALID, FILE_PATH_SDA))
    assert "Error: " in str(e)

    with pytest.raises(ParseException) as e:
        sys_block.StableWrites(context_wrap(STABLE_WRITES_EMPTY))
    assert "Error: " in str(e)


def test_stable_writes_doc_examples():
    env = {
        'block_stable_writes': sys_block.StableWrites(context_wrap(STABLE_WRITES, FILE_PATH_SDA)),
    }
    failed, total = doctest.testmod(sys_block, globs=env)
    assert failed == 0



import doctest
import pytest

from insights.core.exceptions import SkipComponent
from insights.parsers import ipsec_conf
from insights.parsers.ipsec_conf import IpsecConf
from insights.tests import context_wrap


IPSEC_NORMAL = """
# /etc/ipsec.conf - Libreswan IPsec configuration file
#
# see 'man ipsec.conf' and 'man pluto' for more information
#
# For example configurations and documentation, see https://libreswan.org/wiki/

config setup
        # plutodebug="control parsing"
        # plutodebug="all crypt"
        plutodebug=none
        # It seems that T-Mobile in the US and Rogers/Fido in Canada are
        # using 25/8 as "private" address space on their wireless networks.
        # This range has never been announced via BGP (at least up to 2015)
        virtual_private=%v4:10.0.0.0/8,%v4:192.168.0.0/16,%v4:172.16.0.0/12,%v4:25.0.0.0/8,%v4:100.64.0.0/10,%v6:fd00::/8,%v6:fe80::/10

# if it exists, include system wide crypto-policy defaults
include /etc/crypto-policies/back-ends/libreswan.config

# It is best to add your IPsec connections as separate files in /etc/ipsec.d/
include /etc/ipsec.d/*.conf
"""


def test_config_no_data():
    with pytest.raises(SkipComponent):
        IpsecConf(context_wrap(""))


def test_config_dnssec():
    conf = IpsecConf(context_wrap(IPSEC_NORMAL))
    assert len(conf.get('include')) == 2
    assert conf.get('include')[0] == '/etc/crypto-policies/back-ends/libreswan.config'
    assert conf.get('include')[1] == '/etc/ipsec.d/*.conf'
    assert conf.get('config') is not None
    assert conf.get('config').get('setup') is not None
    assert conf['config']['setup']['plutodebug'] == 'none'
    assert conf['config']['setup']['virtual_private'] == '%v4:10.0.0.0/8,%v4:192.168.0.0/16,%v4:172.16.0.0/12,%v4:25.0.0.0/8,%v4:100.64.0.0/10,%v6:fd00::/8,%v6:fe80::/10'
    assert conf.get('conn') is None


def test_doc_examples():
    env = {
        "ipsec_conf": IpsecConf(context_wrap(IPSEC_NORMAL)),
    }
    failed, total = doctest.testmod(ipsec_conf, globs=env)
    assert failed == 0



from insights.parsers import software_collections_list as parsermodule
from insights.parsers.software_collections_list import SoftwareCollectionsListInstalled
from insights.tests import context_wrap
import doctest


EXAMPLE_IN_DOCS = """
devtoolset-7
httpd24
python27
rh-mysql57
rh-nodejs8
rh-php71
rh-python36
rh-ruby24
"""

EXAMPLE_WITH_EMPTY_LINES = """
devtoolset-7
httpd24

python27
rh-mysql57

rh-nodejs8
rh-php71
rh-python36

rh-ruby24
"""

EXAMPLE_WITH_SPACES = """
        devtoolset-7  \t
httpd24
python27   \t
rh-mysql57   \t
  rh-nodejs8  \t
rh-php71
rh-python36
rh-ruby24   \n
"""

EXAMPLE_EMPTY = ''

COLLECTIONS_LIST = [
    'devtoolset-7',
    'httpd24',
    'python27',
    'rh-mysql57',
    'rh-nodejs8',
    'rh-php71',
    'rh-python36',
    'rh-ruby24',
]


def _assert_collections(coll_parser, coll_list):
    assert coll_parser.records == coll_list
    for coll in coll_list:
        assert coll_parser.exists(coll)


def test_module_documentation():
    failed, total = doctest.testmod(parsermodule, globs={
        "collections": SoftwareCollectionsListInstalled(context_wrap(EXAMPLE_IN_DOCS))
    })
    assert failed == 0


def test_swcol_parsing():

    scls = SoftwareCollectionsListInstalled(context_wrap(EXAMPLE_IN_DOCS))
    _assert_collections(scls, COLLECTIONS_LIST)


def test_swcol_exceptions_in_input():
    # Empty lines should be handled properly.
    scls = SoftwareCollectionsListInstalled(context_wrap(EXAMPLE_WITH_EMPTY_LINES))
    _assert_collections(scls, COLLECTIONS_LIST)
    assert not scls.exists('foo')
    # Spaces at beginning and at the end of lines should be handled properly.
    scls = SoftwareCollectionsListInstalled(context_wrap(EXAMPLE_WITH_SPACES))
    _assert_collections(scls, COLLECTIONS_LIST)
    assert not scls.exists('bar')
    # With empty input parser does not contain any records.
    scls = SoftwareCollectionsListInstalled(context_wrap(EXAMPLE_EMPTY))
    _assert_collections(scls, [])
    assert not scls.exists('rh-joke')
